# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BERT finetuning runner with TF-Hub."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import optimization
import run_classifier
import tokenization
import tensorflow as tf
import tensorflow_hub as hub

flags = tf.flags

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "bert_hub_module_handle", None,
    "Handle for the BERT TF-Hub module.")


def create_model(is_training, input_ids, input_mask, segment_ids, labels,
                 num_labels, bert_hub_module_handle):
  """Creates a classification model."""
  tags = set()
  if is_training:
    tags.add("train")
  bert_module = hub.Module(bert_hub_module_handle, tags=tags, trainable=True)
  bert_inputs = dict(
      input_ids=input_ids,
      input_mask=input_mask,
      segment_ids=segment_ids)
  bert_outputs = bert_module(
      inputs=bert_inputs,
      signature="tokens",
      as_dict=True)

  # In the demo, we are doing a simple classification task on the entire
  # segment.
  #
  # If you want to use the token-level output, use
  # bert_outputs["sequence_output"] instead.
  output_layer = bert_outputs["pooled_output"]

  hidden_size = output_layer.shape[-1].value

  output_weights = tf.get_variable(
      "output_weights", [num_labels, hidden_size],
      initializer=tf.truncated_normal_initializer(stddev=0.02))

  output_bias = tf.get_variable(
      "output_bias", [num_labels], initializer=tf.zeros_initializer())

  with tf.variable_scope("loss"):
    if is_training:
      # I.e., 0.1 dropout
      output_layer = tf.nn.dropout(output_layer, keep_prob=0.9)

    logits = tf.matmul(output_layer, output_weights, transpose_b=True)
    logits = tf.nn.bias_add(logits, output_bias)
    probabilities = tf.nn.softmax(logits, axis=-1)
    log_probs = tf.nn.log_softmax(logits, axis=-1)

    one_hot_labels = tf.one_hot(labels, depth=num_labels, dtype=tf.float32)

    per_example_loss = -tf.reduce_sum(one_hot_labels * log_probs, axis=-1)
    loss = tf.reduce_mean(per_example_loss)

    return (loss, per_example_loss, logits, probabilities)


def model_fn_builder(num_labels, learning_rate, num_train_steps,
                     num_warmup_steps, use_tpu, bert_hub_module_handle):
  """Returns `model_fn` closure for TPUEstimator."""

  def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
    """The `model_fn` for TPUEstimator."""

    tf.logging.info("*** Features ***")
    for name in sorted(features.keys()):
      tf.logging.info("  name = %s, shape = %s" % (name, features[name].shape))

    input_ids = features["input_ids"]
    input_mask = features["input_mask"]
    segment_ids = features["segment_ids"]
    label_ids = features["label_ids"]

    is_training = (mode == tf.estimator.ModeKeys.TRAIN)

    (total_loss, per_example_loss, logits, probabilities) = create_model(
        is_training, input_ids, input_mask, segment_ids, label_ids, num_labels,
        bert_hub_module_handle)

    output_spec = None
    if mode == tf.estimator.ModeKeys.TRAIN:
      train_op = optimization.create_optimizer(
          total_loss, learning_rate, num_train_steps, num_warmup_steps, use_tpu)

      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          train_op=train_op)
    elif mode == tf.estimator.ModeKeys.EVAL:

      def metric_fn(per_example_loss, label_ids, logits):
        predictions = tf.argmax(logits, axis=-1, output_type=tf.int32)
        accuracy = tf.metrics.accuracy(label_ids, predictions)
        loss = tf.metrics.mean(per_example_loss)
        return {
            "eval_accuracy": accuracy,
            "eval_loss": loss,
        }

      eval_metrics = (metric_fn, [per_example_loss, label_ids, logits])
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          eval_metrics=eval_metrics)
    elif mode == tf.estimator.ModeKeys.PREDICT:
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode, predictions={"probabilities": probabilities})
    else:
      raise ValueError(
          "Only TRAIN, EVAL and PREDICT modes are supported: %s" % (mode))

    return output_spec

  return model_fn


def create_tokenizer_from_hub_module(bert_hub_module_handle):
  """Get the vocab file and casing info from the Hub module."""
  with tf.Graph().as_default():
    bert_module = hub.Module(bert_hub_module_handle)
    tokenization_info = bert_module(signature="tokenization_info", as_dict=True)
    with tf.Session() as sess:
      vocab_file, do_lower_case = sess.run([tokenization_info["vocab_file"],
                                            tokenization_info["do_lower_case"]])
  return tokenization.FullTokenizer(
      vocab_file=vocab_file, do_lower_case=do_lower_case)


def main(_):
  tf.logging.set_verbosity(tf.logging.INFO)

  processors = {
      "cola": run_classifier.ColaProcessor,
      "mnli": run_classifier.MnliProcessor,
      "mrpc": run_classifier.MrpcProcessor,
  }

  if not FLAGS.do_train and not FLAGS.do_eval:
    raise ValueError("At least one of `do_train` or `do_eval` must be True.")

  tf.gfile.MakeDirs(FLAGS.output_dir)

  task_name = FLAGS.task_name.lower()

  if task_name not in processors:
    raise ValueError("Task not found: %s" % (task_name))

  processor = processors[task_name]()

  label_list = processor.get_labels()

  tokenizer = create_tokenizer_from_hub_module(FLAGS.bert_hub_module_handle)

  tpu_cluster_resolver = None
  if FLAGS.use_tpu and FLAGS.tpu_name:
    tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
        FLAGS.tpu_name, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)

  is_per_host = tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
  run_config = tf.contrib.tpu.RunConfig(
      cluster=tpu_cluster_resolver,
      master=FLAGS.master,
      model_dir=FLAGS.output_dir,
      save_checkpoints_steps=FLAGS.save_checkpoints_steps,
      tpu_config=tf.contrib.tpu.TPUConfig(
          iterations_per_loop=FLAGS.iterations_per_loop,
          num_shards=FLAGS.num_tpu_cores,
          per_host_input_for_training=is_per_host))

  train_examples = None
  num_train_steps = None
  num_warmup_steps = None
  if FLAGS.do_train:
    train_examples = processor.get_train_examples(FLAGS.data_dir)
    num_train_steps = int(
        len(train_examples) / FLAGS.train_batch_size * FLAGS.num_train_epochs)
    num_warmup_steps = int(num_train_steps * FLAGS.warmup_proportion)

  model_fn = model_fn_builder(
      num_labels=len(label_list),
      learning_rate=FLAGS.learning_rate,
      num_train_steps=num_train_steps,
      num_warmup_steps=num_warmup_steps,
      use_tpu=FLAGS.use_tpu,
      bert_hub_module_handle=FLAGS.bert_hub_module_handle)

  # If TPU is not available, this will fall back to normal Estimator on CPU
  # or GPU.
  estimator = tf.contrib.tpu.TPUEstimator(
      use_tpu=FLAGS.use_tpu,
      model_fn=model_fn,
      config=run_config,
      train_batch_size=FLAGS.train_batch_size,
      eval_batch_size=FLAGS.eval_batch_size,
      predict_batch_size=FLAGS.predict_batch_size)

  if FLAGS.do_train:
    train_features = run_classifier.convert_examples_to_features(
        train_examples, label_list, FLAGS.max_seq_length, tokenizer)
    tf.logging.info("***** Running training *****")
    tf.logging.info("  Num examples = %d", len(train_examples))
    tf.logging.info("  Batch size = %d", FLAGS.train_batch_size)
    tf.logging.info("  Num steps = %d", num_train_steps)
    train_input_fn = run_classifier.input_fn_builder(
        features=train_features,
        seq_length=FLAGS.max_seq_length,
        is_training=True,
        drop_remainder=True)
    estimator.train(input_fn=train_input_fn, max_steps=num_train_steps)

  if FLAGS.do_eval:
    eval_examples = processor.get_dev_examples(FLAGS.data_dir)
    eval_features = run_classifier.convert_examples_to_features(
        eval_examples, label_list, FLAGS.max_seq_length, tokenizer)

    tf.logging.info("***** Running evaluation *****")
    tf.logging.info("  Num examples = %d", len(eval_examples))
    tf.logging.info("  Batch size = %d", FLAGS.eval_batch_size)

    # This tells the estimator to run through the entire set.
    eval_steps = None
    # However, if running eval on the TPU, you will need to specify the
    # number of steps.
    if FLAGS.use_tpu:
      # Eval will be slightly WRONG on the TPU because it will truncate
      # the last batch.
      eval_steps = int(len(eval_examples) / FLAGS.eval_batch_size)

    eval_drop_remainder = True if FLAGS.use_tpu else False
    eval_input_fn = run_classifier.input_fn_builder(
        features=eval_features,
        seq_length=FLAGS.max_seq_length,
        is_training=False,
        drop_remainder=eval_drop_remainder)

    result = estimator.evaluate(input_fn=eval_input_fn, steps=eval_steps)

    output_eval_file = os.path.join(FLAGS.output_dir, "eval_results.txt")
    with tf.gfile.GFile(output_eval_file, "w") as writer:
      tf.logging.info("***** Eval results *****")
      for key in sorted(result.keys()):
        tf.logging.info("  %s = %s", key, str(result[key]))
        writer.write("%s = %s\n" % (key, str(result[key])))

  if FLAGS.do_predict:
    predict_examples = processor.get_test_examples(FLAGS.data_dir)
    if FLAGS.use_tpu:
      # Discard batch remainder if running on TPU
      n = len(predict_examples)
      predict_examples = predict_examples[:(n - n % FLAGS.predict_batch_size)]

    predict_file = os.path.join(FLAGS.output_dir, "predict.tf_record")
    run_classifier.file_based_convert_examples_to_features(
        predict_examples, label_list, FLAGS.max_seq_length, tokenizer,
        predict_file)

    tf.logging.info("***** Running prediction*****")
    tf.logging.info("  Num examples = %d", len(predict_examples))
    tf.logging.info("  Batch size = %d", FLAGS.predict_batch_size)

    predict_input_fn = run_classifier.file_based_input_fn_builder(
        input_file=predict_file,
        seq_length=FLAGS.max_seq_length,
        is_training=False,
        drop_remainder=FLAGS.use_tpu)

    result = estimator.predict(input_fn=predict_input_fn)

    output_predict_file = os.path.join(FLAGS.output_dir, "test_results.tsv")
    with tf.gfile.GFile(output_predict_file, "w") as writer:
      tf.logging.info("***** Predict results *****")
      for prediction in result:
        probabilities = prediction["probabilities"]
        output_line = "\t".join(
            str(class_probability)
            for class_probability in probabilities) + "\n"
        writer.write(output_line)


if __name__ == "__main__":
  flags.mark_flag_as_required("data_dir")
  flags.mark_flag_as_required("task_name")
  flags.mark_flag_as_required("bert_hub_module_handle")
  flags.mark_flag_as_required("output_dir")
  tf.app.run()



import numpy as np, logging, sys
import tensorflow as tf
from mautil.tf_layers import EMB, RNN
from mautil.tf_models import TF, InputFeature
from sklearn.model_selection import train_test_split


class Model(TF):
    def _init_input_features(self):
        features = []
        features.append(InputFeature('inputs', [None, 5, 16],tf.float32))
        features.append(InputFeature('inputs_len', [None],tf.float32))
        features.append(InputFeature('tgt', [None, 16],tf.float32))
        return features

    def _add_main_graph(self):
        dims  = [16,16]
        rnn = RNN(dims, bidirection = False)
        _, outs = rnn(self.inputs, inputs_lens = self.inputs_len)
        self._loss = tf.losses.mean_squared_error(outs[-1], self.tgt)

def main():
    train_num = 900
    x = np.random.rand(1000,5, 16)
    y = np.sum(x, -2)
    x, xV, y, yV = train_test_split(x, y, test_size = 0.1, shuffle=False)

    x = {'inputs':x, 'inputs_len':np.ones([len(x)])}
    xV = {'inputs':xV, 'inputs_len':np.ones(len(xV))}
    y = {'tgt':y}
    yV = {'tgt':yV}
    model = Model()
    loss, val_loss = model.fit(x, y, xV, yV)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s:%(threadName)s %(message)s')
    main()



import numpy as np, logging, sys
import mautil
import tensorflow as tf
from mautil.tf_models import TF, InputFeature
from mautil.tf_layers import Dense
from mautil.training import Trainer
from sklearn.model_selection import train_test_split
from mautil import TrainArgParser


class Model(TF):
    def _init_input_features(self):
        features = []
        features.append(InputFeature('inputs', [None, 2],tf.float32)) 
        features.append(InputFeature('tgt', [None],tf.float32))
        return features
        
    def _add_main_graph(self):
        outs = self.inputs
        for dim in [16,16,1]:
            outs = Dense(dim, activation=tf.nn.relu)(outs)
        outs = tf.squeeze(outs, -1)
        self._loss = tf.losses.mean_squared_error(outs, self.tgt)

def gen_data():
    x = np.random.rand(1000,2)
    y = x[:,0] * 2 + x[:,1] * 3 + 1 
    x, xV, y, yV = train_test_split(x, y, test_size = 0.1, shuffle=False)

    x = {'inputs':x}
    xV = {'inputs':xV}
    y = {'tgt':y}
    yV = {'tgt':yV}
    data = {'train':(x, y), 'validate':(xV, yV)}
    return data



if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s:%(threadName)s %(message)s')
    trainer = Trainer('test_model', 9527)
    args = TrainArgParser.load_args()
    if not args.method_or_model:
        args.model_names='Model'
    data = gen_data()
    trainer.train_model(data, args, globals(), process_data=False)



import os, sys
import tensorflow as tf
from ..tf_net.bert.tokenization import BasicTokenizer, load_vocab
from ..tf_net.bert import modeling
from .tf_models import TF, InputFeature


class BERT(TF):
    cfg = TF.cfg.copy()
    cfg.batch_size = 16

    tokenizer = BasicTokenizer()

    def __init__(self, name=None, cfg={}, batch_reader=None):
        super(BERT, self).__init__(name, cfg, batch_reader)
        self.vocab = load_vocab(self.cfg.bert_vocab_file)
        self.cfg.bert_config_file = os.path.join(cfg.data_dir, "bert/chinese_L-12_H-768_A-12/bert_config.json")
        self.cfg.bert_vocab_file = os.path.join(cfg.data_dir, "bert/chinese_L-12_H-768_A-12/vocab.txt")
        self.cfg.init_checkpoint = os.path.join(cfg.data_dir, "bert/chinese_L-12_H-768_A-12/bert_model.ckpt")

    def _init_input_features(self):
        input_features = list()
        input_features.append(InputFeature('seqs', [None, None], tf.int32))
        input_features.append(InputFeature('seqs_len', [None], tf.int32))
        return input_features

    def _add_main_graph(self):
        self.add_bert(self.seqs, self.seqs_len, self._training_plh)

    def add_bert(self, seqs, seqs_len, training):
        mask = tf.cast(tf.sequence_mask(seqs_len, tf.shape(self.seqs)[1]), seqs.dtype)
        bert_config = modeling.BertConfig.from_json_file(self.cfg.bert_config_file)
        self._bert_model = modeling.BertModel(config=bert_config, is_training=training, input_ids=seqs,
                                              input_mask=mask,
                                              use_one_hot_embeddings=False)
        self._feas = self._bert_model.get_sequence_output()
        self._var_init = tf.global_variables_initializer()
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        self.init_var()

    def init_var(self):
        tvars = tf.trainable_variables()
        #initialized_variable_names = {}
        (assignment_map, initialized_variable_names) = modeling.get_assignment_map_from_checkpoint(tvars, self.cfg.init_checkpoint)
        tf.train.init_from_checkpoint(self.cfg.init_checkpoint, assignment_map)
        #with self._sess.graph.as_default():
        #    saver = tf.train.Saver()
        #    saver.restore(self._sess, self.cfg.init_checkpoint)

    def extract_feas(self, ori_batch, batch_size=16):
        num = len(ori_batch['sents'])
        inds = np.arange(num)
        num_batch = (num + batch_size - 1) // batch_size
        feas = [];
        ori_sent_id = 0
        for i in range(num_batch):
            batch = {}
            batch_inds = inds[batch_size * i:min(batch_size * (i + 1), num)]
            batch['batch_inds'] = batch_inds
            batch['sent_ids'] = batch_inds
            batch['sents'] = ori_batch['sents'][batch_inds]
            batch = self.bert_batch(batch)
            fea = self.run(self._sess, batch, self._feas)
            pre_sent_id = -1;
            comb_fea = []
            for f, ind, sent_id in zip(fea, batch['ori_to_tok_inds'], batch['sent_ids']):
                if sent_id != pre_sent_id and len(comb_fea) > 0:
                    feas.append(np.concatenate(comb_fea))
                    comb_fea = []
                comb_fea.append(f[ind])
                pre_sent_id = sent_id
            if len(comb_fea) > 0:
                feas.append(np.concatenate(comb_fea))

        return feas

    def seq(self, chars):
        seq = []
        for char in chars:
            if char in self.vocab:
                seq.append(self.vocab[char])
            else:
                seq.append(self.vocab['[UNK]'])
        return seq

    def bert_sents(self, sents, sents_id, max_len=512):
        ori_to_tok_inds = []
        seqs = []; seqs_len = []; new_sents_ids = []; ori_seqs_len = []
        for sent_id, sent in  enumerate(sents_id, sents):
            ori_to_tok_ind = []
            new_tokens = ['[CLS]']
            ind = 1
            for i, token in enumerate(sent):
                ori_to_tok_ind.append(ind)
                for new_token in self.tokenizer.tokenize(token):
                    new_tokens.append(new_token)
                    ind += 1
                    if ind >= (max_len - 2):
                        new_tokens.append('[SEP]')
                        ori_seqs_len.append(len(ori_to_tok_ind))
                        ori_to_tok_inds.append(ori_to_tok_ind)
                        seqs.append(self.seq(new_tokens))
                        seqs_len.append(len(new_tokens))
                        new_sents_ids.append(sent_id)
                        new_tokens = ['[CLS]']
                        ind = 1
                        ori_to_tok_ind = []
            if len(new_tokens) > 1:
                new_tokens.append('[SEP]')
                ori_seqs_len.append(len(ori_to_tok_ind))
                ori_to_tok_inds.append(ori_to_tok_ind)
                seqs.append(self.seq(new_tokens) + [0] * (max_len-len(new_tokens)))
                seqs_len.append(len(new_tokens))
                new_sents_ids.append(sent_id)
        max_ori_seqs_len = max(ori_seqs_len)
        for ori_to_tok_ind, ori_seq_len in zip(ori_to_tok_inds, ori_seqs_len):
            ori_to_tok_ind + [-1]*(max_ori_seqs_len-ori_seq_len)
        return seqs, seqs_len, ori_seqs_len, new_sents_ids, ori_to_tok_inds



import os, sys, argparse, time, json, re, logging
import numpy as np
import pickle
from contextlib import contextmanager
from collections import OrderedDict, defaultdict

epsilon = 1e-15

def multiple_replace(dict, text):
    # Create a regular expression  from the dictionary keys
    regex = re.compile("|".join(map(re.escape, dict.keys())))
    # For each match, look-up corresponding value in dictionary
    return regex.sub(lambda mo: dict[mo.string[mo.start():mo.end()]], text) 

def load_dump(fpath):
    with open(fpath, 'rb') as f:
        data = pickle.load(f)
    return data


def dump(data, fpath, protocol=2):
    with open(fpath, 'wb') as f:
        pickle.dump(data, f)


def load_json(fpath):
    with open(fpath) as f:
        dictionary = json.load(f)
    return dictionary


def dump_json(dictionary, fpath):
    with open(fpath, 'w') as f:
        json.dump(dictionary, f)


def pad_sequence(seqs, pad=0, seqs_len=None, dtype=np.int32):
    seqs = np.array(seqs)
    shape = seqs.shape
    if seqs.dtype == object:
        flatten_seqs = seqs.flatten()
        seqs_len = np.reshape([len(x) for x in flatten_seqs], shape)
        seqs = np.concatenate(flatten_seqs)
        max_seq_len = np.max(seqs_len)
        mask = np.arange(max_seq_len) < np.expand_dims(seqs_len, -1)
        padded_seqs = np.ones(mask.shape, dtype) * pad
        padded_seqs[mask] = seqs
    else:
        padded_seqs = seqs
        seqs_len = np.ones(shape[:-1], dtype)*shape[-1]
    return padded_seqs, seqs_len


def get_class(kls):
    parts = kls.split('.')
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


def beam_top(scores, n):
    """

    :param scores: [batch_size, beam_size, N]
    :param n: fetch top n
    :return:
    """
    batch_size, beam_size, N = scores.shape
    batch_inds = np.arange(batch_size)
    scores = np.reshape(scores, [batch_size, -1])  # [batch_size, beam_size*N]
    inds = np.argpartition(scores, -n, -1)[:, -n:]
    scores = scores[batch_inds[:, None], inds]
    sort_ind = np.argsort(scores, -1)
    scores = scores[batch_inds[:, None], sort_ind]
    inds = inds[batch_inds[:, None], sort_ind]  # [batch_size, n]
    beam_ids = inds // N
    inds = inds % N
    return inds, beam_ids, scores


def beam_search(loop_func, beam_size, max_len, init_inputs, init_inds, stop=None, return_top=True):
    """

    :param loop_func
            inputs: next_input,  next_inds [batch_size]
            return: score[batch_size, N], next_input

    :param batch_size:
    :param beam_size:
    :param init_inputs:
    :param max_len:
    :param stop: EOS id
    :return: seqs:[batch_size, beam_size, max_len], lens:[batch_size, beam_size]
    """
    batch_size = init_inputs.shape[0]
    batch_inds = np.arange(batch_size)
    beam_next_inputs = np.stack([init_inputs] * beam_size, 1)
    beam_next_inds = np.stack([init_inds] * beam_size, 1)
    lens = np.zeros([batch_size, beam_size], dtype=np.int32)
    beam_scores = None
    inds = []
    beam_ids = []
    beam_stops = np.zeros([batch_size, beam_size], dtype=np.bool)

    for step in range(max_len):
        scores = [];
        next_inputs = []
        for i in range(beam_size):
            score, next_input = loop_func(beam_next_inputs[:,i], beam_next_inds[:, i])  # score:[batch_size, N], next_input:[batch_size, ::]
            scores.append(score)
            next_inputs.append(next_input)
            #if step==0:  # init
            #    break
        scores = np.stack(scores, 1)  # [batch_size, beam_size, N]
        next_inputs = np.stack(next_inputs, 1)  # [batch_size, beam_size, ::]

        N = scores.shape[-1]
        if beam_scores is None:
            #beam_scores = np.zeros([batch_size, beam_size, N])
            beam_scores = np.zeros(scores.shape)
        scores += beam_scores
        step_inds, step_beam_ids, step_beam_scores = beam_top(scores, beam_size)  # [batch_size, beam_size]
        beam_scores = np.expand_dims(step_beam_scores, -1)
        beam_next_inputs = next_inputs[batch_inds[:, None], step_beam_ids]
        beam_next_inds = step_inds
        if stop is not None:
            beam_stops = beam_stops[batch_inds[:, None], step_beam_ids]
            beam_stops = np.logical_or(beam_stops, step_inds == stop)
            if np.all(beam_stops):
                break
            lens = lens[batch_inds[:, None], step_beam_ids]
            lens[~beam_stops] += 1
        inds.append(step_inds)
        beam_ids.append(step_beam_ids)

    seqs = beam_get_seqs(inds, beam_ids, batch_inds)
    if return_top:
        seqs = seqs[:, -1, :]
        lens = lens[:, -1]
    return seqs, lens

def beam_get_seqs(inds, beam_ids, batch_inds):
    inds = inds[::-1]; beam_ids = beam_ids[::-1]
    seqs = [inds[0]]
    for i in range(len(inds)-1):
        seqs.append(inds[i+1][batch_inds[:, None], beam_ids[i]])
        beam_ids[i+1] = beam_ids[i+1][batch_inds[:, None], beam_ids[i]]
    seqs = seqs[::-1]
    seqs = np.stack(seqs, -1)

    return seqs


def timestamp():
    return time.strftime('%Y%m%d%H%M%S')

@contextmanager
def timer(name):
    t0 = time.time()
    logging.info('%s start', name) 
    yield
    logging.info('%s done in %s seconds', name, time.time() - t0) 


class ArgParser(object):
    @staticmethod
    def add_args(parser):
        parser.add_argument("-m", "--method_name", help="run method name")

    @classmethod
    def load_args(cls):
        parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),
                                         formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__)
        cls.add_args(parser)
        args = parser.parse_args()
        return args


def parse_model_name(model_names):
    return model_names.replace(' ', '').split(',')


def merge_args(args, cfg={}):
    for k, v in args.__dict__.items():
        if v is not None:
            cfg[k] = v
    return cfg


def create_model(name, args, cls, seed=None):
    if args.suffix is not None:
        name = name + '_' + args.suffix
    cfg = {}
    if seed is not None:
        cfg['seed'] = seed
    cfg = merge_args(args, cfg)
    model = cls(name=name, cfg=cfg)
    return model


class TrainArgParser(ArgParser):

    @staticmethod
    def add_args(parser):
        parser.add_argument("-m", "--method_or_model", help="run method or model name")
        parser.add_argument("-data_dir", "--data_dir", help="data dir")
        parser.add_argument("-scoring", "--scoring", action="store_true", help="run score after train")
        parser.add_argument("-nv", "--no_validate", action="store_true", help="do not do validation")
        parser.add_argument("-ov", "--only_validate", action="store_true", help="only do validation")
        parser.add_argument("-nt", "--no_train", action="store_true", help="do not do train")
        parser.add_argument("-ntest", "--no_test", action="store_true", help="do not pred test")
        parser.add_argument("-predicting", "--predicting", action="store_true", help="run predict after train")
        parser.add_argument("-dataset", "--dataset", help="dataset")
        parser.add_argument("-brcfg", "--batch_reader_cfg", type=json.loads, help="data reader cfg")
        parser.add_argument("-d", "--debug", action="store_true", help="debug")
        parser.add_argument("-tf_loglevel", "--tf_loglevel", type=int, default=1, help="tf log level.0:DEBUG, 1:INFO, 2:WARN, 3: ERROR")
        parser.add_argument("-use_tf_estimator", "--use_tf_estimator", action="store_true", help="use tensorflow estimator api")
        parser.add_argument("-gsurl", "--gsurl", help="google cloud storage url for save tfrecords and model, used for tpu")
        parser.add_argument("-regen_tfrecord", "--regen_tfrecord", action="store_true", help="regeneration of tf records")
        parser.add_argument("-fake_colab", "--fake_colab", action="store_true", help="used for debug at local")
        parser.add_argument("-warm_start_path", "--warm_start_path", help="tpu warm start")
        parser.add_argument("-use_tpu", "--use_tpu", action="store_true", help="use tpu")
        parser.add_argument("-tpu_data_repeat", type=int, help="when generate tf records, repeat the data")
        parser.add_argument("-tpu_name", "--tpu_name",  help="tpu name")
        parser.add_argument("-num_core_per_host", type=int,  help="num_core_per_host")
        parser.add_argument("-tpu_zone", "--tpu_zone",  help="tpu zone")
        parser.add_argument("-tpu_loop_iterations", type=int,  help="tpu loop iterations")
        parser.add_argument("-gcp_project", "--gcp_project",  help="gcp project")
        parser.add_argument("-save", "--save_model", action="store_true", help="save model after train")
        parser.add_argument("-save_log", action="store_true", help="save log to file")
        parser.add_argument("-save_processed", "--save_processed", action="store_true", help="save processed data")
        parser.add_argument("-save_epoch", "--save_epoch", type=int, help="save per epochs")
        parser.add_argument("-epoch_step", "--epoch_step", type=int, help="steps per epoch")
        parser.add_argument("-save_step", "--save_step", type=int, help="save per steps")
        parser.add_argument("-save_summary_step", "--save_summary_step", type=int, help="save per steps")
        parser.add_argument("-save_keep", "--save_keep", type=int, help="keep number of saved copy")
        parser.add_argument("-lr_decay_step", "--lr_decay_step", type=int, help="lr decay steps")
        parser.add_argument("-gradient_clip", type=float, help="global norm gradient clip")
        parser.add_argument("-warmup_steps", type=int, help="lr warm up steps")
        parser.add_argument("-val_step", "--val_step", type=int, help="do validate per steps")
        parser.add_argument("-val_batchs", "--val_batchs", type=int, help="only run  limited batchs when do validate per steps")
        parser.add_argument("-ckpt_hours", "--keep_checkpoint_every_n_hours", type=int, help="")
        parser.add_argument("-ms", "--model_names", help="model names")
        parser.add_argument("-restore", "--restore", action="store_true", help="save")
        parser.add_argument("-not_restore_cfg", action="store_true", default=False, help="not restore cfg")
        parser.add_argument("-val_test", "--val_test", action="store_true", help="run validation on test split")
        parser.add_argument("-s", "--global_step", help="restore global_step")
        parser.add_argument("-suf", "--suffix", help="suffex model name")
        parser.add_argument("-fprefix", "--fprefix", default='', help="prefix of generated name")
        parser.add_argument("-es", "--es", type=float, help="early stop value")
        parser.add_argument("-ed", "--emb_dim", type=int, help="embedding dim")
        parser.add_argument("-lr", "--lr", type=float, help="learning rate")
        parser.add_argument("-l2", "--l2", type=float, help="l2 regularize")
        parser.add_argument("-bs", "--batch_size", type=int, help="batch size")
        parser.add_argument("-vbs", "--val_batch_size", type=int, help="validate batch size")
        parser.add_argument("-dp", "--dropout", type=float, help="drop out(keep prob)")
        parser.add_argument("-mc", "--min_cnt", type=int, help="minimum count")
        parser.add_argument("-vp", "--vp", type=float, help="valid_percent")
        parser.add_argument("-kfid", "--kfid", help="k fold ID")
        parser.add_argument("-kn", "--kn", type=int, help="k fold num")
        parser.add_argument("-epochs", "--epochs", type=int, help="training epochs")
        parser.add_argument("-tp", "--train_pct", type=float, help="data percent used for train")
        parser.add_argument("-verbose", "--verbose", type=int, help="verbose per num of batch")



import time, logging, numpy as np, json, os
from copy import deepcopy
from collections import namedtuple
from collections import OrderedDict, defaultdict

from mautil.basic_model import BasicModel, InputFeature, LossHist
from mautil.data_reader import create_data_reader
from mautil import tf_util
from mautil import util

import tensorflow as tf

class SessionRunHook(tf.train.SessionRunHook):
    def __init__(self, model):
        self._model = model

    def before_run(self, run_context):
        return self._model.before_run(run_context)

    def after_run(self, run_context, run_values):
        self._model.after_run(runcontext, run_values)


class TF(BasicModel, tf.train.SessionRunHook):
    cfg = deepcopy(BasicModel.cfg)
    cfg.verbose = 30
    cfg.global_step = None
    cfg.batch_reader_cfg = {'process_mode': 'T'}
    cfg.gradient_clip = None
    cfg.batch_reader = 'BatchReader'
    cfg.dropout = None
    cfg.tf_qsize = 4
    cfg.keep_checkpoint_every_n_hours = 1000000000
    cfg.save_keep = 1000000000
    cfg.save_summary_step = 1000000000
    cfg.use_tpu = False
    cfg.tpu_data_repeat = 1
    cfg.tpu_name = None
    cfg.warm_start_path = None  # used by tpu
    cfg.tpu_zone = None
    cfg.gcp_project = None
    cfg.num_core_per_host = 1
    cfg.num_hosts = 1
    cfg.tpu_loop_iterations = 2

    def __init__(self, name = 'TF', cfg={}, batch_reader=None):
        super(TF, self).__init__(name, cfg)
        if self.cfg.use_tpu:
            self.cfg.use_tf_estimator=True
        self._input_features = self._init_input_features()
        if batch_reader is None:
            batch_reader = self.cfg.batch_reader

        if isinstance(batch_reader, str):
            self.batch_reader = create_data_reader(batch_reader, 'batch_reader', self.cfg.seed, self.gen_fname(''), **self.cfg.batch_reader_cfg)
        else:
            self.batch_reader = batch_reader
        self._run_phase = 'train'
        self._loss_hist = None
        tf_util.set_tfloglevel(self.cfg.tf_loglevel)
        if self.cfg.use_tpu:
            assert self.cfg.batch_size%self.cfg.num_core_per_host == 0
            assert self.cfg.val_batch_size%self.cfg.num_core_per_host == 0


    def begin(self):
        self._loss_hist = LossHist()
        self._run_step = 0

    def after_create_session(self, session, coord):
        is_training = session.run(self._training_plh)
        if is_training:
            self._run_phase = 'train'
        elif self._run_phase != 'predict':
            self._run_phase = 'validate'

    def before_run(self, run_context):
        self._run_step += 1
        if self._run_phase == 'train':
            fetches = self.train_nodes

        elif self._run_phase == 'validate':
            fetches = self.validate_nodes

        elif self._run_phase == 'predict':
            fetches = self._pred_dict
        else:
            raise NotImplemented
        run_args = tf.train.SessionRunArgs(fetches=fetches, feed_dict={})
        return run_args

    def after_run(self, run_context, run_values):
        loss_hist = self._loss_hist
        outs = run_values.results
        lr = outs.get('lr', 'NA')
        gnorm = outs.get('gnorm', 'NA')
        loss_hist.append(outs['losses'])
        global_step = outs['global_step']

        if self._run_step % self.cfg.verbose == 0:
            avg_loss_str = loss_hist.avg_output()
            logging.info('  name:%s, phase %s, global step:%s, lr:%s, gnorm:%s, loss is:%s', self.name, self._run_phase, global_step, lr,gnorm, avg_loss_str)
        if global_step % self.cfg.save_step == 0 and not self.cfg.use_tf_estimator:
            self.save(global_step)
        if self._run_step % self.cfg.epoch_step == 0 and self._run_phase == 'train':
            avg_loss_str = loss_hist.avg_output()
            logging.info('name:%s, phase %s, epoch:%s, global step:%s, loss is:%s', self.name, self._run_phase, global_step//self.cfg.epoch_step, global_step, avg_loss_str)
            self._loss_hist = LossHist()


    def destroy(self):
        super(TF, self).destroy()
        tf.reset_default_graph()

    def _init_input_features(self):
        return []

    def create_model(self):
        self._graph = tf.Graph()
        self.sess = tf.Session(graph = self._graph)
        self._pred_dict = OrderedDict()
        with self._graph.as_default():
            tf.set_random_seed(self.cfg.seed)
            self._create_graph()
            self.saver = tf.train.Saver(max_to_keep=self.cfg.save_keep, keep_checkpoint_every_n_hours=self.cfg.keep_checkpoint_every_n_hours)
        super(TF, self).create_model()

    def _add_predict(self):
        pass

    def _add_main_graph(self):
        pass

    def _create_graph(self):
        self._add_plh()
        self._add_main_graph()
        self._add_predict()
        self._add_loss()
        self._add_train()
        self._add_train_nodes()
        self._init_vars()



    def _add_loss(self):
        pass

    def _add_train(self):
        self._var_list = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES)
        self._train_op, self._gradients, self._global_step, self._lr_node, self._opt, self._gnorm = self._add_train_op(self._loss, self._var_list, clip=self.cfg.gradient_clip)

    def _add_train_nodes(self):
        """
        the name loss is preversed. The fit_epoch function will use key "loss" to fetch the train loss or validation loss,
        if the model have loss of different source, you can add them in the train_nodes and validate_nodes to monitor them.
        For example, the model have loss_1 and loss_2, loss=loss_1+loss_2, you can add these 3 nodes in the validate_nodes.
        :return:
        """
        n_params = int(sum([np.prod(v.shape) for v in tf.trainable_variables()]))
        logging.info('total num of trainable params:%s', n_params)

        train_losses = OrderedDict()
        train_losses['loss'] = self._loss
        self.train_nodes = {'losses': train_losses, 'global_step': self._global_step, 'lr':self._lr_node, 'train_op': self._train_op}
        if self._gnorm is not None:
            self.train_nodes['gnorm'] = self._gnorm

        validate_losses = OrderedDict()
        validate_losses['loss'] = self._loss
        self.validate_nodes = {'losses': validate_losses, 'global_step': self._global_step}

    def _init_vars(self):
        self._var_init = tf.global_variables_initializer()
        self.sess.run(self._var_init)

    def _add_opt(self, loss, var_list, name='train_op', clip=None, lr_node=None, global_step=None, opt=None, opt_name='adam'):
        if global_step is None:
            global_step = tf.train.get_global_step()
            if global_step is None:
                global_step = tf.Variable(0, dtype=tf.int64, trainable=False, name='global_step')

        if self.cfg.warmup_steps > 0:
            warmup_lr = tf.to_float(global_step+1) / tf.to_float(self.cfg.warmup_steps) * self.cfg.lr
        else:
            warmup_lr = 0.0
        if lr_node is None:
            #decaly_lr = tf.train.exponential_decay(learning_rate=self._lr_plh, global_step=global_step-self.cfg.warmup_steps, staircase=False, decay_steps = self.cfg.lr_decay_step, decay_rate=self.cfg.lr_decay_rate, name = 'learning_rate')
            decay_lr = tf.train.cosine_decay(self._lr_plh, global_step=global_step - self.cfg.warmup_steps, decay_steps=self.cfg.lr_decay_step, alpha=self.cfg.min_lr_ratio)
            lr_node = tf.where(global_step < self.cfg.warmup_steps, warmup_lr, decay_lr)
        if opt is None:
            opt = tf.train.AdamOptimizer(learning_rate=lr_node, name=opt_name)
            if self.cfg.use_tpu:
                opt = tf.contrib.tpu.CrossShardOptimizer(opt)
        return opt, lr_node, global_step

    def _add_train_op(self, loss, var_list, name='train_op', clip=None, lr_node=None, global_step=None, opt=None, opt_name='adam'):
        opt, lr_node, global_step = self._add_opt(loss, var_list, name=name, clip=clip, lr_node=lr_node, global_step=global_step, opt=opt, opt_name=opt_name)

        grads = tf.gradients(loss, var_list)
        gnorm = None
        #self._add_summary([loss])
        #self._add_summary([grad for grad, var in gradients if grad is not None], tf.summary.histogram)
        if clip is not None:
            (grads, gnorm) = tf.clip_by_global_norm(grads, clip_norm=clip)
        #if cfg.batch_norm:
        #    update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        #    with tf.control_dependencies(update_ops):
        #        self._train_op = opt.apply_gradients(clipped, global_step=self._global_step, name = 'train_op')
        #else:
        train_op = opt.apply_gradients(zip(grads, var_list), global_step=global_step, name=name)
        #self._train_op = opt.minimize(self.loss, global_step=self._global_step, name = 'train_op')
        return train_op, grads, global_step, lr_node, opt, gnorm

    def _add_inputs_plh(self, input_features):
        input_plhs = {}
        for fea in input_features:
            if fea.sparse:
                plh = tf.sparse.placeholder(fea.dtype, fea.shape, name=fea.name)
            else:
                plh = tf.placeholder(fea.dtype, fea.shape, name=fea.name)
            setattr(self, fea.name, plh)
            input_plhs[fea.name] = plh
        return input_plhs

#        if self.cfg.use_tf_estimator:
#            dtypes, shapes = zip(*[[fea.dtype, fea.shape] for fea in self._input_features])
#            self._tf_queue = tf.FIFOQueue(capacity=self.cfg.tf_qsize, dtypes=dtypes, shapes=shapes)
#            self.enqueue_op = self._tf_queue.enqueue(self._input_plhs)
#            self._input_tensors = self._tf_queue.dequeue()
#            for tensor in self._input_tensors:
#                setattr(self, tensor.name, tensor)

    def _add_plh(self):
        self._lr_plh = tf.placeholder(tf.float32, name = "lr_plh" )
        self._training_plh = tf.placeholder(tf.bool, name="training_plh" )
        if self.cfg.dropout is not None:
            self._dropout_plh = tf.placeholder(tf.float32, None, name="dropout_plh" )
        else:
            self._dropout_plh = None
        self._input_plhs = self._add_inputs_plh(self._input_features)

    def _add_emb(self, inputs, size, dim = None, name='emb', embedding = None, trainable=True, scope = 'emb'):
        initializer = tf.contrib.layers.xavier_initializer(uniform=True)
        with tf.variable_scope(scope, initializer = initializer, reuse = tf.AUTO_REUSE):
            if embedding is None:
                embed_var = tf.get_variable(name + '_W', [size, dim], tf.float32, initializer = initializer, trainable = trainable)
            else:
                embed_var = tf.get_variable(name + '_W', initializer = embedding, trainable = trainable)
        embeded = tf.nn.embedding_lookup(embed_var, inputs, name = name)
        return embeded, embed_var


    def save(self, global_step = None):
        super(TF, self).save()
        fpath = self.gen_fname('', 'model.ckpt')
        save_path = self.saver.save(self.sess, fpath, global_step=global_step)
        logging.info("Model saved to file:{}".format(save_path))
        self.batch_reader.save()

    def get_checkpoint_path(self, model_dir=None, global_step=None):
        if model_dir is None:
            model_dir = self.gen_fname('')
        if global_step is None:
            ckpt = tf.train.get_checkpoint_state(model_dir)
            model_path = ckpt.model_checkpoint_path
        else:
            model_path = os.path.join(model_dir, 'model.ckpt-{}'.format(global_step))
        return model_path


    def restore(self, global_step=None, var_list=None, model_dir=None):
        if global_step is None:
            global_step = self.cfg.global_step  # keep it before the restore of cfg
        if not self.cfg.not_restore_cfg:
            super(TF, self).restore()
        if self._model is None:
            self.create_model()

        model_path = self.get_checkpoint_path(model_dir, global_step)
        with self.sess.graph.as_default():
            saver = tf.train.Saver(var_list=var_list)
            saver.restore(self.sess, model_path)
            logging.info("Model restored from file:{}".format(model_path))
        if not self.cfg.only_validate and not self.cfg.no_train:
            self.batch_reader.restore()

    def get_feed(self, batch):
        feed = {
                }
        if 'lr' in batch:
            feed[self._lr_plh] = batch['lr']
            feed[self._training_plh] = True
        else:
            feed[self._training_plh] = False

        if 'dropout' in batch:
            feed[self._dropout_plh] = batch['dropout']

        for fea in self._input_features:
            if fea.name in batch:
                plh = getattr(self, fea.name)
                feed[plh] = batch[fea.name]

        return feed

    def run(self, sess, batch, nodes):
        feed = self.get_feed(batch)
        outputs = sess.run(nodes, feed)
        return outputs

    def _get_lr(self, itr):
        return self.cfg.lr

    def pre_run_batch(self, batch, epoch=0, itr=0, global_step=0, is_training=True):
        if is_training:
            batch['lr'] = self._get_lr(itr)
            if self.cfg.dropout is not None:
                batch['dropout'] = self.cfg.dropout
        else:
            if self.cfg.dropout is not None:
                batch['dropout'] = 1.0
        return batch

    def _get_tfrecord_spec(self, input_features):
        record_spec = {}
        for fea in input_features:
            record_spec[fea.name] = tf.FixedLenFeature(fea.shape, fea.dtype)
        return record_spec

    def _get_input_fn(self, x, y, shuffle, data_type):
        record_spec = self._get_tfrecord_spec(self._input_features)

        def _parse_tfrecord(tfrecord):
            examples = tf.parse_single_example(tfrecord, record_spec)
            return examples

        n_batch = None
        if self.cfg.use_tpu:
            fname = self.name + '_' + data_type + '.tfrecords'
            info_fname = self.name + '_' + data_type + '.tfrecords_info'
            record_info = {}
            gsurl = os.path.join(self.cfg.gsurl, os.path.basename(self.gen_fname('')), fname)
            info_gsurl = os.path.join(self.cfg.gsurl, os.path.basename(self.gen_fname('')), info_fname)
            fpath = self.gen_fname('', fname)
            info_fpath = self.gen_fname('', info_fname)
            cmd = 'gsutil -q stat ' + gsurl
            rst = os.system(cmd)
            if rst != 0 or self.cfg.regen_tfrecord:
                if not os.path.exists(fpath) or self.cfg.regen_tfrecord:
                    if data_type == 'train':
                        repeat = self.cfg.tpu_data_repeat
                    else:
                        repeat = 1
                    ds = self._create_dataset(x, y, self.cfg.batch_size, shuffle=shuffle, data_type=data_type, repeat=repeat)
                    n_batch = tf_util.ds2tfrecord(ds, self._input_features, fpath)
                    record_info['n_batch'] = n_batch
                    logging.info('tf records file generated:%s, total of %s batch', fpath, n_batch)
                    util.dump_json(record_info, info_fpath)
                if not self.cfg.fake_colab:
                    cmd = 'gsutil cp ' + info_fpath + ' ' + info_gsurl
                    rst1 = os.system(cmd)
                    logging.info('file %s copied to %s', info_fpath, info_gsurl)
                    assert rst1 == 0, cmd
                    cmd = 'gsutil cp ' + fpath + ' ' + gsurl
                    rst2 = os.system(cmd)
                    assert rst2 == 0, cmd
                    logging.info('file %s copied to %s', fpath, gsurl)
            else:
                record_info = util.load_json(info_fpath)
                n_batch = record_info['n_batch']

        def _input_fn(params):
            batch_size = params.get('batch_size')
            logging.info('input func batch size is %s', batch_size)
            if not self.cfg.use_tpu:
                ds = self._create_dataset(x, y, batch_size, data_type=data_type, shuffle=shuffle)
            else:
                if not self.cfg.fake_colab:
                    ds = tf.data.TFRecordDataset(gsurl)
                else:
                    ds = tf.data.TFRecordDataset(fpath)
                ds = ds.map(_parse_tfrecord).apply(tf.data.experimental.unbatch())
                if data_type == 'train':
                    ds = ds.cache().repeat()
                    ds = ds.batch(batch_size, drop_remainder=True)
                    ds = ds.prefetch(self.cfg.num_core_per_host * batch_size)
                else:
                    ds = ds.batch(batch_size, drop_remainder=True)
            return ds
        return _input_fn, n_batch

    def _create_dataset(self, x, y, batch_size, data_type, shuffle=False, repeat=1):
        dtypes = {}; shapes = {}
        record_spec = self._get_tfrecord_spec(self._input_features)
        for fea in self._input_features:
            if fea.name in record_spec:
                dtypes[fea.name] = fea.dtype
                shapes[fea.name] = fea.shape

        def _gen():
            for _ in range(repeat):
                for i, batch in enumerate(self.batch_reader(x, batch_size, y, shuffle=shuffle, data_type=data_type)):
                    features = {}
                    for fea in self._input_features:
                        if fea.name in batch:
                            features[fea.name] = batch[fea.name]
                    yield features
        ds = tf.data.Dataset.from_generator(_gen, dtypes, shapes)
        return ds

    def _set_model_fn_inputs(self, features, params):
        for name, fea in features.items():
            setattr(self, name, fea)
            logging.info("feature %s:%s", name, fea)
        plh_features = []
        for fea in self._input_features:
            if fea.name not in features and not self.cfg.use_tpu:
                plh_features.append(fea)
        self._input_plhs = self._add_inputs_plh(plh_features)

    def _metric_fn(self, loss):
        return {
            "loss": tf.metrics.mean(loss),
        }

    def _get_estimator_spec(self, mode, batch_size, hooks=None):
        output_spec = None
        if mode == tf.estimator.ModeKeys.TRAIN:
            output_spec = tf.contrib.tpu.TPUEstimatorSpec(
                mode=mode,
                loss=self._loss,
                train_op=self._train_op,
                training_hooks=hooks,
            )
        elif mode == tf.estimator.ModeKeys.EVAL:
            loss = self._loss
            if self.cfg.use_tpu:
                with tf.colocate_with(loss):
                    loss = tf.contrib.tpu.cross_replica_sum(loss) \
                                 / self.cfg.num_hosts / self.cfg.num_core_per_host
            metric_loss = tf.tile(tf.reshape(loss, [1, 1]), [batch_size, 1])

            output_spec = tf.contrib.tpu.TPUEstimatorSpec(
                mode=mode,
                loss=loss,
                evaluation_hooks=hooks,
                eval_metrics=(self._metric_fn, [metric_loss]),
                )

        return output_spec

    def _get_model_fn(self, ):
        def model_fn(features, labels, mode, params):
            batch_size = params['batch_size']
            logging.info('model func batch_size:%s', batch_size)
            self._set_model_fn_inputs(features, params)

            is_training = (mode == tf.estimator.ModeKeys.TRAIN)
            self._training_plh = tf.constant(is_training)
            if self.cfg.dropout is not None:
                if is_training:
                    self._dropout_plh = tf.constant(self.cfg.dropout)
                else:
                    self._dropout_plh = tf.constant(1.0)
            self._lr_plh = tf.constant(self.cfg.lr)
            self._add_main_graph()
            self._add_loss()
            self._add_train()
            self._add_train_nodes()
            return self._get_estimator_spec(mode, batch_size, hooks=[self])
        return model_fn

    def _get_estimator(self, model_fn, run_config, warm_start_from):
        estimator = tf.contrib.tpu.TPUEstimator(
            use_tpu=self.cfg.use_tpu,
            model_fn=model_fn,
            config=run_config,
            train_batch_size=self.cfg.batch_size,
            eval_batch_size=self.cfg.val_batch_size,
            warm_start_from=warm_start_from)
        return estimator

    def fit_estimator(self, x, y, xV = None, yV = None, save=False, restore=False, xTest=None):
        if self.cfg.fake_colab:
            train_input_fn, n_batch = self._get_input_fn(x, y, shuffle=False, data_type="train")
            val_input_fn, n_batch = self._get_input_fn(xV, yV, shuffle=False, data_type="validate")
            ds = val_input_fn(params={'batch_size':self.cfg.batch_size//self.cfg.num_core_per_host})
            fea_dict = ds.make_one_shot_iterator().get_next()
            with tf.Session() as sess:
                steps = n_batch//self.cfg.batch_size
                try:
                    for i in range(steps):
                        aaa = sess.run(fea_dict)
                except tf.errors.OutOfRangeError as e:
                    logging.error('out of range, n_batch:%s, num:%s', n_batch, num)
                    raise e
                finally:
                    logging.error('fake colab done')


        tpu_cluster_resolver = None
        if self.cfg.use_tpu:
            tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
                self.cfg.tpu_name, zone=self.cfg.tpu_zone, project=self.cfg.gcp_project)
            model_dir = os.path.join(self.cfg.gsurl, os.path.basename(self.gen_fname('')))
        else:
            model_dir = self.gen_fname('')

        is_per_host = tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
        run_config = tf.contrib.tpu.RunConfig(
            cluster=tpu_cluster_resolver,
            model_dir=model_dir,
            session_config=tf.ConfigProto(allow_soft_placement=True, log_device_placement=True),
            save_checkpoints_steps=self.cfg.save_step,
            save_summary_steps=self.cfg.save_summary_step,
            save_checkpoints_secs=None,
            keep_checkpoint_max=self.cfg.save_keep,
            tpu_config=tf.contrib.tpu.TPUConfig(
                num_shards=self.cfg.num_core_per_host * self.cfg.num_hosts,
                iterations_per_loop = self.cfg.tpu_loop_iterations,
                per_host_input_for_training=is_per_host))

        model_fn = self._get_model_fn()

        # If TPU is not available, this will fall back to normal Estimator on CPU
        # or GPU.
        # warm start
        warm_start_from = None
        if self.cfg.use_tpu:
            if self.cfg.warm_start_path is not None:
                warm_start_from = tf.estimator.WarmStartSettings(ckpt_to_initialize_from=self.cfg.warm_start_path)
        else:
            if self.cfg.restore:
                checkpoint_path = self.get_checkpoint_path(global_step=self.cfg.global_step)
                warm_start_from = tf.estimator.WarmStartSettings( ckpt_to_initialize_from=checkpoint_path)

        estimator = self._get_estimator(model_fn, run_config, warm_start_from)

        best_val_loss = np.inf; val_loss = None
        for i in range(self.cfg.epochs):
            if not self.cfg.only_validate:
                train_input_fn, n_batch = self._get_input_fn(x, y, shuffle=False, data_type="train")
                if n_batch is not None:
                    steps = n_batch//self.cfg.num_hosts
                else:
                    steps = 1000000000
                self._run_phase = 'train'
                logging.info('run phase:%s, steps:%s', self._run_phase, steps)
                estimator.train(input_fn=train_input_fn, steps=steps)
                logging.info('train epoch %s done', i)
            if xV is not None and not self.cfg.no_validate:
                eval_input_fn, n_batch = self._get_input_fn(xV, yV, shuffle=False, data_type="validate")
                if n_batch is not None:
                    steps = n_batch
                else:
                    steps = 1000000000
                self._run_phase = 'validate'
                logging.info('run phase:%s, steps:%s', self._run_phase, steps)
                val_loss = estimator.evaluate(input_fn=eval_input_fn, steps=steps)['loss']
                logging.info('val loss for epoch %s is %s', i, val_loss)
                if self._should_stop(best_val_loss, val_loss):
                    break
                if val_loss < best_val_loss:
                    best_val_loss = val_loss


    def fit(self, x, y, xV = None, yV = None, save=False, restore=False, xTest=None, no_validate=False, only_validate=False):
        if self.cfg.use_tf_estimator:
            return self.fit_estimator(x, y, xV=xV, yV=yV, save=save, restore=restore, xTest=xTest)
        else:
            return super(TF, self).fit(x, y, xV=xV, yV=yV, save=save, restore=restore, xTest=xTest)

    def fit_batch(self, batch):
        outs = self.run(self.sess, batch, self.train_nodes)
        return outs

    def _update_loss_hist(self, losses, loss_hist):
        for name, loss in losses.items():
            loss_hist[name].append(loss)

    def _get_avg_loss(self, loss_hist):
        avg_loss = OrderedDict()
        for key, v in loss_hist.items():
            avg_loss[key] = np.mean(v)
        return avg_loss

    def _fit_epoch(self, x, y, xV = None, yV = None, epoch=None, xTest = None):
        loss_hist = LossHist()

        if not self.cfg.only_validate:
            itr = 0
            global_step = 0
            for i, batch in enumerate(self.batch_reader(x, self.cfg.batch_size, y, shuffle=True, data_type='train')):
                batch = self.pre_run_batch(batch, epoch, itr, global_step)
                itr += 1
                outs = self.fit_batch(batch)
                lr = outs['lr']
                gnorm = outs.get('gnorm', 'NA')
                loss_hist.append(outs['losses'])
                global_step = outs['global_step']
                if (i+1) % self.cfg.verbose == 0:
                    avg_loss_str = loss_hist.avg_output()
                    logging.info('  name:%s,global step:%s, lr:%s, gnorm:%s, train loss is:%s, totally %s batchs', self.name, global_step, lr,gnorm, avg_loss_str, i+1)
                if (global_step+1)%self.cfg.save_step == 0:
                    self.save(global_step)
            avg_loss_str = loss_hist.avg_output()
            logging.info('name:%s, epoch:%s, global step:%s,train loss is:%s, totally %s batchs', self.name, epoch, global_step, avg_loss_str, i+1)
            loss = loss_hist.get_avg()['loss']
            if ((epoch+1)%self.cfg.save_epoch) == 0:
                self.save(global_step)
        else:
            global_step = 0
            loss = 0

        if xV is not None and not self.cfg.no_validate:
            val_loss = self.val_epoch(xV, yV, epoch, global_step)
        else:
            val_loss = None
        return loss, val_loss

    def val_epoch(self, xV, yV, epoch, global_step):
        loss_hist = LossHist()
        num_sample = 0
        for i, batch in enumerate(self.batch_reader(xV, self.cfg.val_batch_size, yV, shuffle=False, data_type='validate')):
            batch = self.pre_run_batch(batch, epoch, i, global_step, is_training=False)
            batch_size = batch['batch_size']
            num_sample += batch_size
            outs = self.run(self.sess, batch, self.validate_nodes)
            loss_hist.append(outs['losses'])
        avg_loss_str = loss_hist.avg_output()
        logging.info("epoch:%s, val loss is:%s", epoch, avg_loss_str)
        val_loss = loss_hist.get_avg()['loss']
        return val_loss


    def predict(self, x, pred_dict = None, do_concat=True, to_list=False):
        if pred_dict is None:
            pred_dict = self._pred_dict
        preds = defaultdict(list)
        for i, batch in enumerate(self.batch_reader(x, self.cfg.batch_size, shuffle=False, data_type='validate')):
            batch = self.pre_run_batch(batch, 0, i, i, is_training=False)
            batch_pred = self.run(self.sess, batch, pred_dict)
            for k, v in batch_pred.items():
                if to_list:
                    preds[k].extend(list(v))
                else:
                    preds[k].append(v)
        if do_concat:
            for k in preds:
                preds[k] = np.concatenate(preds[k],0)
        return preds







import unittest
import numpy as np

class Test(unittest.TestCase):
    seed=9527
    def __init__(self, *args, **kwargs):
        super(Test, self).__init__(*args, **kwargs)
        self._rs = np.random.RandomState(self.seed)




from base_test_class import Test
from mautil import util
import numpy as np
import unittest


class TestUtil(Test):
    def gen_score(self, target_inds, N, beam_ids= None):
        batch_size, beam_size, max_len = target_inds.shape
        #scores = np.zeros([batch_size, beam_size, max_len, N])
        scores = []
        #max_scores = np.zeros([batch_size, 1])
        max_scores = np.zeros([batch_size, beam_size])
        if beam_ids is None:
            beam_ids = np.arange(beam_size)
        for step in range(max_len):
            #step_scores = scores[:,:,step,:]
            step_scores = np.zeros([batch_size, beam_size, N])
            inds = target_inds[:, :, step:step+1]
            #beam_scores = max_scores + np.expand_dims(np.arange(beam_size)+1, 0)*0.01
            beam_scores = np.max(max_scores, -1, keepdims=True) + np.expand_dims(np.arange(beam_size)+1, 0)*0.01
            #np.random.shuffle(beam_ids)
            #step_scores[np.arange(batch_size)[:,None,None], np.arange(beam_size)[None,:,None], inds] += np.expand_dims(beam_scores, -1)
            step_scores[np.arange(batch_size)[:,None,None], beam_ids[None,:,None], inds] = np.expand_dims(beam_scores-max_scores[:, beam_ids], -1)
            scores.append(step_scores)
            #max_scores = np.max(beam_scores, -1, keepdims=True)
            max_scores = beam_scores
        return np.stack(scores,2)
        
    def test_beam_search(self):
        N = 6
        inds = [];
        inds.append([[1,1,1],[2,3,5],[0,0,1],[0,3,4]])
        inds.append([[2,2,2],[1,2,4],[1,1,1],[2,1,4]])
        inds = np.stack(inds, 1)
        scores = self.gen_score(inds, N)

        cross_beam_inds = []
        cross_beam_inds.append([[0,2,3]])
        cross_beam_inds.append([[1,1,4]])
        cross_beam_inds = np.stack(cross_beam_inds, 1)
        beam_ids = np.array([0,0])
        cross_beam_scores = self.gen_score(np.array(cross_beam_inds), N, beam_ids)
        cross_beam_inds[-1][-1] = np.array([0,2,4]) # correct it from first beam id
        scores = np.concatenate([scores, cross_beam_scores], 0)
        target_inds = np.concatenate([inds, cross_beam_inds], 0)

        target_lens = np.array([[0, 3],[3, 0],[2, 0],[3, 1],[3, 3]])
        print('scores is {}'.format(scores))
        print('target lens is {}'.format(target_lens))

        batch_size, beam_size, max_len = target_inds.shape
        inputs = np.zeros([batch_size, beam_size, max_len, 10])
        init_inds = np.array([-1] * batch_size)
                    
        stop = 1
        cnt = 0
        def loop_func(inp, inds):
            nonlocal cnt
            next_inputs = inputs[:, cnt%beam_size, cnt//beam_size, :]
            next_scores = scores[:, cnt%beam_size, cnt//beam_size, :]
            cnt += 1
            return next_scores, next_inputs
    
        inds, lens = util.beam_search(loop_func, beam_size, max_len, inputs[:,0,0,:], init_inds, stop=stop, return_top=False)
        print('beam search inds is {}'.format(inds))
        print('beam search lens is {}'.format(lens))
        assert np.all(np.array(target_inds)==inds)
        assert np.all(np.array(target_lens)==lens)
        

if  __name__ == "__main__":
    unittest.main()
    
        
        



import numpy as np
import tensorflow as tf
from . import util


def get_vars(scopes):
    vars = []
    for scope in scopes:
        vars += tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=scope)
    return vars


def gumbel_softmax(logits, gamma, eps=1e-20):
    U = tf.random_uniform(tf.shape(logits))
    G = -tf.log(-tf.log(U + eps) + eps)
    return tf.nn.softmax((logits + G) / gamma)
    #return tf.nn.softmax((logits+eps) / gamma)


def rnn_decode(cell, inputs, loop_func, max_len, initial_state=None):
    outs = []; hs = []
    h = initial_state
    for t in range(max_len):
        cell_out, h = cell(inputs, h)
        hs.append(h)
        inputs, out = loop_func(cell_out)
        outs.append(out)
    hs = tf.stack(hs, 1)
    if isinstance(initial_state, tuple):
        if len(initial_state) == 1:
            hs = tf.squeeze(hs, 0)  # [time, batch_size,::]
            hs = transpose(hs, 0, 1)  # [batch_size, time, ::]

    return hs, tf.stack(outs, 1)


def csr2tfsparse(x):
    coo = x.tocoo()
    indices = np.mat([coo.row, coo.col]).transpose()
    return tf.SparseTensorValue(indices, coo.data, coo.shape)


def restore_vars(sess, checkpoint_path, meta_path):
    saver = tf.train.import_meta_graph(meta_path)
    saver.restore(sess, checkpoint_path)

def tensor_to_feature(tensor):

    return tf.train.Feature(bytes_list=bytes_list)


def _int64_feature(values):
  return tf.train.Feature(int64_list=tf.train.Int64List(value=values))

def _float_feature(values):
  return tf.train.Feature(float_list=tf.train.FloatList(value=values))

def ds2tfrecord(ds, input_features, filepath):
    convert_funcs = {}
    for fea in input_features:
        if fea.dtype  in [tf.int32, tf.int16, tf.int64]:
            convert_funcs[fea.name] = _int64_feature
        elif fea.dtype in [tf.float32, tf.float64]:
            convert_funcs[fea.name] = _float_feature
        else:
            raise NotImplemented

    with tf.python_io.TFRecordWriter(filepath) as writer:
        feat_dict = ds.make_one_shot_iterator().get_next()
        num = 0
        with tf.Session() as sess:
            try:
                while True:
                    features = {}
                    feat_npy = sess.run(feat_dict)
                    for name, npy in feat_npy.items():
                        features[name] = convert_funcs[name](npy.flatten())
                    # Create a Features message using tf.train.Example.
                    example_proto = tf.train.Example(features=tf.train.Features(feature=features))
                    example_string = example_proto.SerializeToString()
                    # Write to TFRecord
                    writer.write(example_string)
                    num += 1
            except tf.errors.OutOfRangeError:
                pass
    return num

def set_tfloglevel(level=1):
    tf_loglevel = {0: tf.logging.DEBUG, 1: tf.logging.INFO, 2: tf.logging.WARN, 3: tf.logging.ERROR}
    tf.logging.set_verbosity(tf_loglevel[level])

def gather_seqs(seqs, gather_inds):
    dtype = gather_inds.dtype
    n = len(gather_inds.get_shape())
    shape = tf.shape(gather_inds)
    r_shape = [shape[0]]
    inds=tf.range(shape[0]); tile=[1]
    inds = tf.expand_dims(inds, -1)
    for i in range(1, n-1):
        inds=tf.expand_dims(inds, i)
        inds = tf.tile(inds, tile+[shape[i]] +[1])
        r_shape.append(shape[i])
        r_inds = tf.range(shape[i])
        r_inds = tf.expand_dims(r_inds, -1)
        for j in tile[::-1]:
            r_inds = tf.expand_dims(r_inds, 0)
        r_inds = r_inds * tf.ones(r_shape+[1], tf.int32)
        inds = tf.concat([inds, r_inds], -1)
        tile += [1]
    inds = tf.tile(tf.expand_dims(inds, -2), tile + [shape[-1]] + [1])
    gather_inds = tf.expand_dims(gather_inds, -1)
    gather_inds = tf.concat([inds, gather_inds], -1)
    rst = tf.gather_nd(seqs, gather_inds)
    return rst


def transpose_nd(tensor, dims1, dims2):
    """
    switch dims1[i] and dims2[i]
    :param tensor:
    :param dims1:
    :param dims2:
    :return:
    """
    assert len(dims1) == len(dims2), 'num of dims must be same'
    assert len(dims1+dims2) == len(set(dims1+dims2)), 'there must have no dup of dim'
    ndims = tensor.shape.ndims
    shape = np.arange(ndims)
    for dim1, dim2 in zip(dims1, dims2):
        dim = shape[dim1]
        shape[dim1] = shape[dim2]
        shape[dim2] = dim
    return tf.transpose(tensor, shape)


def transpose(tensor, dim1, dim2):
    return transpose_nd(tensor, [dim1], [dim2])

def tile(input, dims, repeats):
    ndims = input.shpae.ndims
    shape = [1] * ndims
    for dim, r in zip(dims, repeats):
       shape[dim] = r
    return tf.tile(input, shape)

def expand_tile_like(t1, t2):

    if t1.shape.ndims < t2.shape.ndims:
        src = t1; tgt = t2
    elif t1.shape.ndims > t2.shape.ndims:
        src = t2; tgt = t1
    else:
        return t1, t2

    shape = [1]*tgt.shape.ndims
    tgt_shape = tf.shape(tgt)
    for i in range(tgt.shape.ndims-src.shape.ndims):
        shape[i] = tgt_shape[i]
        src = src[None, ...]
    src = tf.tile(src, shape)

    if t1.shape.ndims < t2.shape.ndims:
        return src, tgt
    else:
        return tgt, src



def create_rnn_cell(dims, cell_cls='tf.nn.rnn_cell.GRUCell', dropout=None):
    """
     
    :param dims: list of cell dim [128, 128]
    :param cell_cls:
    :return:
    """
    cell_cls = util.get_class(cell_cls)
    cells = []
    for dim in dims:
        cell = cell_cls(dim)
        if dropout is not None:
            cell = tf.nn.rnn_cell.DropoutWrapper(cell, input_keep_prob=dropout)
        cells.append(cell)
    multi_cell = tf.nn.rnn_cell.MultiRNNCell(cells, state_is_tuple=True)
    return multi_cell



import numpy as np, time, logging, os, bisect
import scipy
from queue import Queue, Empty
from threading import Thread
import multiprocessing as mp
from copy import deepcopy

from . import util
from . import tf_util

gl = globals()


class BasicBatchReader(object):
    def __init__(self, name, seed, data_dir, batch_process_func=None):
        self.name = name
        self._rs = np.random.RandomState(seed)
        self._batch_process_func = batch_process_func
        self._queue = None
        self.state_fpath = os.path.join(data_dir, name+'.dump')
        self.state = {'rs':self._rs, 'batch_pos':0}

    def set_rs(self, rs):
        self._rs = rs
    
    def save(self):
        state = deepcopy(self.state)
        if self._queue is not None:
            state['batch_pos'] = max(0, state['batch_pos'] - self._queue.qsize())

        util.dump(state, self.state_fpath)
        logging.info('batch reader %s batch position %s saved to file %s', self.name, state['batch_pos'],  self.state_fpath)
        
    def restore(self):
        if os.path.exists(self.state_fpath):
            self.state = util.load_dump(self.state_fpath)
        else:
            logging.warning('batch reader did not find the state file %s to restore, will do initial run', self.state_fpath)
            self.state = {'rs':self._rs, 'batch_pos':0}
        logging.info('batch reader %s current state is %s', self.name, self.state)


def create_data_reader(cls_name, name, seed, data_dir, **kwargs):
    if cls_name in gl:
        cls = gl[cls_name]
    else:
        cls = util.get_class(cls_name)
    return cls(name, seed, data_dir, **kwargs)


class BatchReader(BasicBatchReader):
    def __init__(self, name, seed, data_dir, batch_process_func=None, process_mode='T', sleep_seconds=1, timeout=1, qsize=8, balanced_read=False, squeeze_seq=False, min_seq_len=-1, ordered=False,  **kwargs):
        """

        :param name:
        :param seed:
        :param process_mode: T for thread, P for subprocess, N for neither
        :param sleep_seconds:
        :param timeout:
        :param qsize:
        :param squeeze_seq: seqs will be squeezed to max_seq_len of batch
        """
        super(BatchReader, self).__init__(name, seed, data_dir, batch_process_func)
        self._sleep_seconds = sleep_seconds
        self._process_mode = process_mode
        self._timeout = timeout
        self._qsize = qsize
        self._balanced_read = balanced_read
        self._do_squeeze_seq = squeeze_seq
        self._min_seq_len = min_seq_len
        self._ordered = ordered
        self.kwargs = kwargs

        if self._process_mode == 'P':
            self._queue = mp.Queue(self._qsize)
        elif self._process_mode == 'T':
            self._queue = Queue(self._qsize)


    def _squeeze_seq(self, batch, seq_keys):
        for seq_key, seq_len_key in seq_keys:
            seq_key = '_'.join(seq_len_key.split('_')[:-1])
            max_len = max(self._min_seq_len, np.max(batch[seq_len_key]))
            batch[seq_key] = batch[seq_key][:, :max_len]
        return batch

    def _get_seqs_keys(self, keys):
        seqs_len_keys = []
        seqs_keys = []
        for key in keys:
            if key.endswith('seqs_len'):
                seqs_len_keys.append(key)
                seqs_key = '_'.join(key.split('_')[:-1])
                seqs_keys.append(seqs_key)
        return seqs_keys, seqs_len_keys

    def get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        self.state['rs'] = deepcopy(self._rs)
        batch_pos = self.state['batch_pos']
        if data_type != 'train':
            self.state['batch_pos'] = 0
        seqs_keys, seqs_len_keys = self._get_seqs_keys(x.keys())

        get_batch_func = self._get_batch
        if self._balanced_read:
            get_batch_func = self._get_batch_balanced
        for batch in get_batch_func(x, batch_size, y, shuffle, data_type):
            if self._do_squeeze_seq is True:
                batch = self._squeeze_seq(batch, zip(seqs_keys, seqs_len_keys))

            if self._batch_process_func is not None:
                batch = self._batch_process_func(batch, data_type)
            self.state['batch_pos'] = (self.state['batch_pos'] + 1) % self._num_batch
            for key in batch:
                if isinstance(batch[key], scipy.sparse.csr_matrix):
                    batch[key] = tf_util.csr2tfsparse(batch[key])
            yield batch
        if data_type == 'train':
            self.state['batch_pos'] = 0
        else:
            self.state['batch_pos'] = batch_pos
        self.state['rs'] = self._rs

    def _get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        for key in x:
            num = x[key].shape[0]
            break
        inds = np.arange(num)
        if shuffle and not self._ordered:
            self._rs.shuffle(inds)
        num_batch = (num + batch_size -1)//batch_size
        self._num_batch = num_batch
        logging.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.int32)
        if shuffle:
            self._rs.shuffle(batch_ids)

        for i in batch_ids[self.state['batch_pos']:]:
            batch = {}
            batch_inds = inds[batch_size*i:min(batch_size*(i+1), num)]
            batch['batch_inds'] = batch_inds
            batch['batch_size'] = len(batch_inds)
            for key in x:
                batch[key] = x[key][batch_inds]

            if y is not None:
                for key in y:
                    batch[key] = y[key][batch_inds]
            yield batch

    def _gen_batch(self, x, batch_size, y, shuffle, data_type='train'):
        for i, batch in enumerate(self.get_batch(x, batch_size, y, shuffle=shuffle, data_type=data_type)):
            self._queue.put(batch, block=True)
        self._queue.put(None)

    def __call__(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """

        :param x: dict of array. key name endswith _seqs, _seqs_len were preserved for seqs data
        :param batch_size:
        :param y:
        :param shuffle:
        :param data_type:
        :return:
        """

        if self._process_mode == 'P':
            thread = mp.Process(target=self._gen_batch, args=[x, batch_size, y, shuffle, data_type])
        elif self._process_mode == 'T':
            thread = Thread(target=self._gen_batch, args=[x, batch_size, y, shuffle, data_type])
        if self._process_mode in ['T', 'P']:
            thread.daemon = True
            thread.start()
            time.sleep(self._sleep_seconds)
            logging.info('name:%s, get batch sleep %s seconds', self.name, self._sleep_seconds)
            while True:
                batch = self._queue.get(True, self._timeout)
                if batch is None:break;
                yield batch
        elif self._process_mode == 'N':
            for batch in self.get_batch(x, batch_size, y, shuffle=shuffle, data_type=data_type):
                yield batch


    def _get_batch_balanced(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """
        In some case data is imbalance
        """

        label_inds = []
        labels = np.unique(x['labels'])
        min_num =  np.inf
        for label in labels:
            label_ind = list(np.where(np.array(x['labels']) == label)[0])
            if shuffle and not self._ordered:
                self._rs.shuffle(label_ind)
            label_inds.append(label_ind)
            if len(label_ind) < min_num:
                min_num = len(label_ind)
        num_batch = (min_num + batch_size -1)//batch_size
        self._num_batch = num_batch
        logging.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.int32)
        if shuffle:
            self._rs.shuffle(batch_ids)

        for i in batch_ids[self.state['batch_pos']:]:
            batch = {}
            batch_inds = []
            for inds in label_inds:
                batch_inds += inds[batch_size*i:min(batch_size*(i+1), min_num)]
            batch['batch_inds'] = batch_inds
            batch['batch_size'] = len(batch_inds)
            for key in x:
                batch[key] = x[key][batch_inds]

            if y is not None:
                for key in y:
                    batch[key] = y[key][batch_inds]

            yield batch


class SeqBatchReader(BatchReader):
    def _get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """
        all seqs will be concatenated
        :param x:
        :param batch_size:
        :param y:
        :param shuffle:
        :param data_type:
        :return:
        """
        max_seq_len = self.kwargs.get('max_seq_len')
        seqs = x['seqs'];
        pad = self.kwargs.get('pad', 0)


        num = 0; seqs_len = []
        for seq in seqs:
            l = len(seq)
            seqs_len.append(l)
            num += (l + max_seq_len -1)//max_seq_len
        cumsum = np.cumsum(seqs_len)
        inds = np.arange(num, dtype=np.uint32)
        if shuffle and not self._ordered:
            self._rs.shuffle(inds)
        num_batch = (num + batch_size -1)//batch_size
        self._num_batch = num_batch
        logging.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.uint32)

        for i in batch_ids[self.state['batch_pos']:]:
            batch_seqs = []; batch_seqs_len = []; batch={}
            batch_inds = inds[batch_size*i:min(batch_size*(i+1), num)]
            batch_seq_inds = []
            batch_seq_example_inds = []
            batch['batch_inds'] = batch_inds
            for ind in batch_inds:
                seq_id = bisect.bisect_left(cumsum, ind + 1)
                batch_seq_inds.append(seq_id)
                seq = seqs[seq_id]
                if seq_id == 0:
                    seq_example_ind = ind
                else:
                    seq_example_ind = ind - cumsum(seq_id-1)
                batch_seq_example_inds.append(seq_example_ind)
                batch_seqs.append(seq[seq_example_ind*max_seq_len:(seq_example_ind+1)*max_seq_len])
            batch['seqs'], batch['seqs_len'] = util.pad_sequence(batch_seqs, pad=pad)
            batch['batch_size'] = len(batch_inds)
            batch['batch_seq_inds'] = batch_seq_inds
            batch['batch_seq_example_inds'] = batch_seq_example_inds
            if self._batch_process_func is not None:
                batch = self._batch_process_func(batch, data_type)
            yield batch


class BatchedSeqBatchReader(BatchReader):
    """
    all seqs will be concatenated and divided to batch_size  seqs.
    each example in batch is from each seq respectively
    """
    def _get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """
        :param x:
        :param batch_size:
        :param y:
        :param shuffle:
        :param data_type:
        :return:
        """
        max_seq_len = self.kwargs.get('max_seq_len')
        seqs = np.concatenate(x['seqs'])
        num_batch = len(seqs)//(batch_size*max_seq_len)
        self._num_batch = num_batch
        start_ind = self._rs.choice(len(seqs)%(batch_size*max_seq_len)+1)
        seqs = seqs[start_ind:start_ind + num_batch*batch_size*max_seq_len]
        batchs = np.reshape(seqs, [batch_size, num_batch, max_seq_len])


        logging.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.uint32)
        inds = np.arange(num_batch*batch_size)

        for i in range(self.state['batch_pos'], num_batch):
            batch = {}
            batch_inds = inds[batch_size*i:batch_size*(i+1)]
            batch['batch_inds'] = batch_inds
            batch['seqs'] = batchs[:, i, :]
            batch['seqs_len'] = np.ones([batch_size], np.int32) * max_seq_len
            batch['batch_size'] = batch_size
            if self._batch_process_func is not None:
                batch = self._batch_process_func(batch, data_type)
            yield batch



import json, os, logging, pprint
import numpy as np
from copy import deepcopy
from collections import namedtuple, defaultdict, OrderedDict
from . import util
from .text_process import Vocabulary

InputFeature = namedtuple('InputFeature', ['name', 'shape', 'dtype', 'sparse'])
InputFeature.__new__.__defaults__ = (None,) * len(InputFeature._fields)


def copy_dict(src, dst):
    dic = {}
    for w in dst:
        dic[w] = src[w]
    return dic

class BasicCFG(object):
    def update(self, cfg):
        for k, v in cfg.items():
            attr = getattr(self, k, None)
            if isinstance(attr, dict):
                attr.update(v)
            else:
                setattr(self, k, v)

    def save(self, fpath):
        with open(fpath, 'w') as f:
            json.dump(self.__dict__, f)

    def load(self, fpath):
        with open(fpath) as f:
            cfg = json.load(f)
        self.update(cfg)

    def copy(self):
        return deepcopy(self)

    def __str__(self):
        return pprint.pformat(self.__dict__, indent=4)


class CFG(BasicCFG):
    def __init__(self):
        self.seed = 9798
        self.lr = 1e-3
        self.min_lr_ratio = 0.001
        self.warmup_steps = 1000
        self.es = -0.001
        self.batch_size = 32
        self.val_batch_size = self.batch_size
        self.lr_decay_step = 1000000000
        self.save_step = 1000000000
        self.save_epoch = 1000000000
        self.epoch_step = 1000000000
        self.val_step = 1000000000
        self.val_batchs = 1000000000
        self.emb_dim = 8
        self.epochs = 10
        self.scoring = False
        self.min_cnt = 5 # minimum count for building vocabulary
        self.save_model = False  # if save model after fit
        self.only_validate = False
        self.no_validate = False
        self.debug = False
        self.dropout = None
        self.data_dir = 'data'
        self.vocab_file = None
        self.tokenizer = None
        self.copy_source = False  # if copy the model definition source file when save model


class LossHist(defaultdict):
    def __init__(self, decimal_digits=8, *args, **kwargs):
        self._fstr = "{}:{:." + str(8) + "f}"
        super(LossHist, self).__init__(list, *args, **kwargs)

    def append(self, losses):
        for name, loss in losses.items():
            self[name].append(loss)

    def get_avg(self):
        avg_loss = OrderedDict()
        keys = sorted(self.keys())
        for key in keys:
            avg_loss[key] = np.mean(self[key])
        return avg_loss

    def get_avg_sum(self):
        avg_loss = self.get_avg()
        return np.sum([v for k, v in avg_loss.items()])

    def avg_output(self):
        avg_loss = self.get_avg()
        #tot = np.sum(v for k, v in avg_loss.items())
        #avg_loss['tot'] = tot
        outstr = ', '.join(self._fstr.format(k, v) for k, v in avg_loss.items())
        return outstr


class CallBack(object):

    def __init__(self, itr=0):
        self.itr = itr

    def on_train_begin(self):
        pass

    def on_batch_start(self, batch):
        pass

    def on_batch_end(self, epoch):
        self.itr += 1


class BasicModel(object):
    cfg = CFG()
    vocab = None
       
    def __init__(self, name, cfg={}):
        self.name = name
        self.update_cfg = cfg
        self.cfg = deepcopy(self.cfg)
        self.cfg.update(self.update_cfg)

        if self.cfg.vocab_file is not None:
            if self.cfg.debug:
                self.cfg.vocab_file = self.cfg.vocab_file + '_debug'
#            self.vocab = Vocabulary('vocab', os.path.abspath(self.cfg.vocab_file))
            self.vocab = util.get_class(self.cfg.vocab)('vocab', self.cfg.vocab_file, min_cnt=self.cfg.min_cnt)
            self.cfg.vocab_size = self.vocab.size

#        if self.name is None:
#            self.name = self.__class__.__name__
        self._rs = np.random.RandomState(self.cfg.seed)

        self._model = None

        dirname = self.gen_fname('')
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        script_file = os.path.basename(__file__)
        if self.cfg.save_model:
            if not os.path.exists(self.cfg.data_dir):
                os.mkdir(self.cfg.data_dir)
            fpath = self.gen_fname('', script_file)
            dirname = os.path.dirname(fpath)
            if self.cfg.copy_source:
                shutil.copy2(script_file, dirname)
        if self.cfg.save_log:
            flog = self.gen_fname('', util.timestamp()+'.log')
            file_handler = logging.FileHandler(flog)
            log_formatter = logging.Formatter('%(asctime)s %(levelname)s:%(threadName)s %(message)s')
            file_handler.setFormatter(log_formatter)
            logging.getLogger().addHandler(file_handler)

        logging.info('model cfg is: %s', self.cfg)

    def destroy(self):
        """
        release resource
        """
        pass

    def create_model(self):
        self._model = self

    def _process_data(self, data, data_type):
        """
        process data and update configuration like cfg.emb_size if needed
        :param data:
        :param data_type: 'train', 'validate', 'test'
        :return: processed_data, label
        """
        return data

    def process_data(self, data_dict):
        """
        :param data_dict: dictionary of data. keys:train, validate, test
        :return:
        """
        data = []
        for data_type in ['train', 'validate', 'test']:  # keep train first
            if data_type in data_dict:
                with util.timer('process data type {}'.format(data_type)):
                    processed_data = list(self._process_data(data_dict[data_type], data_type))
                for d in processed_data:
                    if d is not None:
                        for k in d:
                            if isinstance(d[k], list):
                                d[k] = np.array(d[k])
                data += processed_data
            else:
                data += [None, None]
        return data

    def fit(self, x, y, xV = None, yV = None, save=False, restore=False, xTest=None):
        if self._model is None:
            self.create_model()
        if restore:
            self.restore()

        best_val_loss = np.inf
        for i in range(self.cfg.epochs):
            logging.info("fit start")
            loss, val_loss = self._fit_epoch(x, y, xV, yV, epoch=i, xTest=xTest)
            if not self.cfg.no_validate:
                if self.cfg.scoring:
                    score = self.score(xV, yV)
                if not loss <np.inf:
                    logging.error('something is wrong loss is:%s, exit', loss)
                    raise Exception('loss:{} is abnormal'.format(loss))

                if self._should_stop(best_val_loss, val_loss):
                    break
                else:
                    if best_val_loss > val_loss:
                        best_val_loss = val_loss

        if save and self.cfg.save_epoch != 1:
            self.save()

        logging.info('***** end fit')
        return loss, best_val_loss

    def _should_stop(self, best_val_loss, val_loss):
        stop = False
        if (best_val_loss-val_loss) / val_loss < -self.cfg.es:
            logging.info("(best_val_loss %s - val_loss %s) / val_loss %s < minus early stop %s, train done", best_val_loss, val_loss, val_loss, self.cfg.es)
            stop = True
        return stop


    def gen_fname(self, postfix, *paras):
        name = self.name
        if self.cfg.debug:
            name = 'debug_' + name
        if postfix:
            name += '_' + postfix
        fname = os.path.join(self.cfg.data_dir, name, *paras)
        return fname
    
    def restore(self):
        fpath = self.gen_fname('', 'cfg.json')
        self.cfg.load(fpath)
        self.cfg.update(self.update_cfg)
        logging.info('model cfg after restore is: %s', self.cfg)


    def save(self):
        fpath = self.gen_fname('', 'cfg.json')
        dirname = os.path.dirname(fpath)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        self.cfg.save(fpath)


    def save_predict(self, pred, suffix=''):
        fpath = self.gen_fname('', 'pred' + suffix + '.dump')
        util.dump(pred, fpath)

        
    def predict(self, x):
        pred = self._model.predict(x)
        return pred

    def score(self, x, y):
        pass



import os, math, time
from .util import timer
from glob import glob
import logging
from collections import OrderedDict


gl = globals()


class DataSets(object):
    _registered = {'wikitext103':'WikiText103',
                 }

    def __init__(self, data_dir='data/datasets'):
        self._data_dir = data_dir

    def load_dataset(self, dataset, **kwargs):
        ds_dir = kwargs.pop('data_dir', os.path.join(self._data_dir, dataset))
        if dataset in self._registered:
            ds = gl[self._registered[dataset]](dataset, data_dir=ds_dir)
        else:
            ds = gl['DataSet'](dataset, data_dir=ds_dir)
        data_dict = ds.load_data(**kwargs)
        return data_dict


class DataSet(object):
    _data_type = None
    _splits = ['train', 'valid', 'test']
    _download_cmds = {}

    def __init__(self, name, data_dir, data_type=None):
        self.name = name
        self._data_dir = data_dir
        if data_type is not None:
            self._data_type = data_type

    def load_plaintext_file(self, fpath, use_line=True, num_line=math.inf):
        with open(fpath) as f:
            if use_line:
                data = []
                for i, line in enumerate(f):
                    data.append(line)
                    if i>num_line:
                        break
            else:
                data = f.read()
        return data

    def load_plaintext(self, fdir, splits, **kwargs):
        use_line = kwargs.pop('use_line', True)
        num_line = kwargs.pop('num_line', math.inf)
        data_dict = {}

        for split, reg in splits.items():
            split_data = {}
            for fpath in glob(os.path.join(fdir, '**', reg + '*'), recursive=True):
                fdata = self.load_plaintext_file(fpath, use_line, num_line)
                split_data[fpath] = fdata
            assert len(split_data)>0, 'please check the split:{}, exists in folder:{}'.format(reg, fdir)
            data_dict[split] = split_data
            logging.info('first 5 files for split:%s are::%s', split, list(data_dict[split].keys())[0:5])
        return data_dict

    def download(self, ds_dir):
        if not os.path.exists(ds_dir):
            rst = os.system('mkdir -p {}'.format(ds_dir))
            assert rst==0

        if not glob(os.path.join(ds_dir, '**', '*'+ self._splits[0] + '*'), recursive=True):
            error_msg = "download datset {} failed to {}".format(self.name, ds_dir)
            try:
                for fname, cmd in self._download_cmds.items():
                    with timer('download dataset {} for file {}'.format(self.name, fname)):
                        rst = os.system(cmd.format(ds_dir))
                assert rst==0, error_msg
            except Exception as e:
                time.sleep(2)
                logging.error('download failed, you can manually put the train, valid, test files in folder %s', ds_dir)
                raise(e)

    def load_data(self, **kwargs):
        debug = kwargs.get('debug', False)
        ds_dir = kwargs.pop('ds_dir', self._data_dir)
        datatype = kwargs.pop('datatype')
        splits = kwargs.pop('splits', self._splits)
        if isinstance(splits, list):
            splits = {split: split for split in splits}
        if debug:
            for k, v in splits.items():
                splits[k] = 'debug_' + v

        if datatype == 'plaintext':
            data_dict = self.load_plaintext(ds_dir, splits, **kwargs)

        for split in list(data_dict.keys()):
            if 'valid' == split:
                data_dict['validate'] = data_dict[split]
                _ = data_dict.pop(split)

        return data_dict


class RegisteredDataSet(DataSet):
    def load_data(self, **kwargs):
        _ = kwargs.pop('datatype', None)
        self.download(self._data_dir)
        data_dict = super(RegistedDataSet, self).load_data(data_type=self._data_type, splits=splits, **kwargs)
        return data_dict


class WikiText103(RegisteredDataSet):
    _download_cmds = {'wikitext-103-v1.zip': 'wget --continue https://s3.amazonaws.com/research.metamind.io/wikitext/wikitext-103-v1.zip -O {}.zip \
                                   && unzip -q {}.zip -d {}',
                 }





import numpy as np, logging, time, traceback, gc
from . import util
from .util import timer


class Trainer(object):
    def __init__(self, name, seed):
        self.name = name
        self.seed = seed

    def train_model(self, data, args, gl, process_data=True, copies=1, sleep_seconds=1):
        """

        :param data: dictionary of train, validate, test data.
        :param args: ArgParser args
        :param gl: dictionary of model classes
        :param process_data: True if the data has not been processed by model
        :param copies: num of copies to train for each model
        #param sleep_seconds: wait seconds after each copy trained(sometimes gpu resource was not released quickly)
        :return:
        """
        rs = np.random.RandomState(self.seed)
        preds_test = []
        preds_validate = []
        names = util.parse_model_name(args.model_names)
        if args.no_validate:
            _ = data.pop('validate', None)
        if args.no_test:
            _ = data.pop('test', None)

        d_train = data.get('train')
        d_validate = data.get('validate', None)
        d_test = data.get('test', None)

        for i, name in enumerate(names):
            logging.info('will train for model: %s', name)
            processed = False
            for j in range(copies):
                if copies == 1:
                    seed = None
                else:
                    seed = rs.randint(np.power(2, 32))
                model = util.create_model(name, args, gl[name], seed)
                if process_data and not processed:
                    with timer('process data'):
                        x_train, y_train, x_validate, y_validate, x_test, y_test =  \
                            model.process_data(data)
                    processed = True
                else:
                    x_train, y_train = d_train
                    if d_validate is not None:
                        x_validate, y_validate = d_validate
                    else:
                        x_validate = None
                        y_validate = None
                    if d_test is not None:
                        x_test, y_test = d_test
                    else:
                        x_test = None
                if args.no_train:
                    model.create_model()
                    model.restore()
                    break
                try:
                    model.fit(x_train, y_train, x_validate, y_validate, args.save_model, xTest=x_test, restore=args.restore)
                    break
                except Exception as e:
                    traceback.print_exc()
                    logging.error('********error when fit %sth  model %s:%s********', i, name, e)
                    model.destroy()
                    del model
                    gc.collect()
                finally:
                    time.sleep(sleep_seconds)
                    logging.info('sleeped 1 second for %sth model  %s', i, name)
            if args.val_test and x_test is not None:
                logging.info('run validation on test')
                model.val_epoch(x_test, y_test, epoch=0, global_step=0)
            logging.info('train done for modle:%s,start predict', name)
            if args.predicting:
                suffix = ('_' + str(args.global_step)) if args.global_step is not None else ''
                if x_validate is not None:
                    pred_validate = model.predict(x_validate)
                    model.save_predict(pred_validate, suffix)
                    preds_validate.append(pred_validate)
                if x_test is not None:
                    pred_test = model.predict(x_test)
                    preds_test.append(pred_test)
                    if args.save_model:
                        model.save_predict(pred_test, suffix+'_test')

            model.destroy()
            del model
            gc.collect()
        if args.predicting and not args.no_validate:
            if copies>1:
                pred = np.mean(preds_validate, 0)
            else:
                pred = preds_validate[0]
            return pred



import os, logging
from collections import Counter
from . import util


class Vocabulary(object):
    fake = {'[PAD]': 0, '[SOS]': 1, '[EOS]': 2, '[UNK]': 3}

    def __init__(self, name, vocab_path, min_cnt=5, tokenizer=None):
        self.name = name
        self.vocab_path = vocab_path
        self.size = None
        self.vocab = None
        self.tokenizer = tokenizer
        self.min_cnt = min_cnt

        if os.path.isfile(self.vocab_path):
            self.restore()

    def restore(self):
        logging.info('load vocab from saved file: %s', self.vocab_path)
        self.vocab = util.load_json(self.vocab_path)
        self.size = len(self.vocab['word2id'])

    def _build_vocab(self, cnt):
        word2id = self.fake.copy()
        id2word = [k for k, v in sorted(word2id.items(), key=lambda x:x[1])]
        for word in cnt:
            if cnt[word] >= self.min_cnt:
                word2id[word] = len(word2id)
                id2word.append(word)
        self.size = len(word2id)
        self.vocab = {'word2id': word2id, 'id2word': id2word, 'word_cnt': dict(cnt)}

    def build_vocab(self, sents, save=True):
        if self.vocab is None:
            logging.info('building vocab')

            cnt = Counter()
            if self.tokenizer:
                for sent in sents:
                    for word in self.tokenizer(sent):
                        cnt.update([word])
            else:
                for sent in sents:
                    for word in sent:
                        cnt.update([word])
            self._build_vocab(cnt)
            if save:
                util.dump_json(self.vocab, self.vocab_path)
                logging.info('vocab saved to :%s', self.vocab_path)
        logging.info('vocab size:%s', self.size)

    def seq(self, sents, max_len=None, add_sos=True, add_eos=True):
        w2i = self.vocab['word2id']
        pad = w2i['[PAD]']
        add_len = 0
        if add_sos:
            add_len += 1
        if add_eos:
            add_len += 1
        seqs = []
        seq_lens = []
        for sent in sents:
            if self.tokenizer:
                tokens = self.tokenizer(sent)
            else:
                tokens = sent
            inds = [w2i[w] if w in w2i else w2i['[UNK]'] for w in tokens]
            if max_len is not None:
                inds = inds[0:max_len-add_len]
            if add_sos:
                inds = [w2i['[SOS]']] + inds
            if add_eos:
                inds += [w2i['[EOS]']]
            seq_lens.append(len(inds))
            if max_len is not None:
                inds += [pad] * (max_len - len(inds))
            seqs.append(inds)
        return seqs, seq_lens

    def deseq(self, seqs):
        id2w = self.vocab['id2word']
        sents = []
        for seq in seqs:
            sent = []
            for word_id in seq:
                sent.append(id2w[word_id])
            sents.append(sent)
        return sents






"""
the layer implementation for training variables sharing is based completely on scope.
This means if you create 2 layer object of Dense with same scope name, they will share variables.
By default it will use the layer name as the scope name
"""
import inspect
from collections import namedtuple
from functools import partial
from . import tf_util

import tensorflow as tf
from tensorflow.python.util import nest
from tensorflow.python.framework import tensor_shape
from tensorflow.python.framework import ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import control_flow_ops


def add_dropout(inputs, dropout):
    outputs = []
    if not isinstance(inputs, tuple):
        inputs = tuple([inputs])

    for inp in inputs:
        if isinstance(dropout, Layer):
            output = dropout(inp)
        else:
            output = tf.nn.dropout(inp, keep_prob=dropout)
        outputs.append(output)
    if len(outputs) == 1:
        outputs = outputs[0]
    else:
        outputs = tuple(outputs)
    return outputs


class Layer(object):
    uid = 0

    def __init__(self, name = None, scope = None, initializer = None, **kwargs):
        self.name = self.__class__.__name__ + '_' + str(self.uid)
        self.__class__.uid += 1
        if name is not None:
            self.name = name

        if initializer is not None:
            self._initializer = initializer
        else:
            self._initializer = self._create_default_initializer()

        if scope is not None:
            with tf.variable_scope(scope, initializer=self._initializer, reuse=tf.AUTO_REUSE) as scope:
                self.scope = scope
        else:
            with tf.variable_scope(self.name, initializer=self._initializer, reuse=tf.AUTO_REUSE) as scope:
                self.scope = scope

        self._kwargs = kwargs
        self._input_dropout = kwargs.get('input_dropout', None)
        self._output_dropout = kwargs.get('output_dropout', None)

    def _create_default_initializer(self):
        return tf.contrib.layers.xavier_initializer(uniform=False)

    def get_vars(self):
        return tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope.name)

    @property
    def weights(self):
        return self.get_vars()

    def _create_graph(self, inputs):
        return None

    def __call__(self, inputs, *args, **kwargs):
        with tf.variable_scope(self.scope, initializer=self._initializer):
            if self._input_dropout is not None:
                inputs = add_dropout(inputs, self._input_dropout)
            outputs = self._create_graph(inputs, *args, **kwargs)
            if self._output_dropout is not None:
                outputs = add_dropout(outputs, self._output_dropout)
            return outputs

    def __rrshift__(self, other, *args, **kwargs):
        """ >> """
        return self.__call__(other, *args, **kwargs)


class Dense(Layer):
    uid = 0

    def __init__(self, dim, activation=None, name=None, scope=None, initializer=None, **kwargs):
        super(Dense, self).__init__(name, scope, initializer, **kwargs)
        self.dim = dim
        self.activation = activation
        self.dense = partial(tf.layers.dense, kernel_initializer=self._initializer)

    def _create_graph(self, inputs):
        outs = self.dense(inputs, self.dim, activation=self.activation)
        return outs


class EMB(Layer):
    uid = 0

    def __init__(self, size, dim = None, embedding = None, trainable = True, name = None, scope=None, initializer=None, dtype=tf.float32, **kwargs):
        super(EMB, self).__init__(name, scope, initializer, **kwargs)
        self.size = size
        self.dim = dim
        self.embedding = embedding
        self.trainable = trainable
        self.dtype = dtype

    def _create_graph(self, inputs):
        if self.embedding is None:
            embed_var = tf.get_variable(self.name, [self.size, self.dim], dtype=self.dtype, trainable=self.trainable)
        else:
            embed_var = tf.get_variable(self.name, initializer=self.embedding, trainable=self.trainable)
        embedded = tf.nn.embedding_lookup(embed_var, inputs)
        return embedded


class CNN(Layer):
    uid = 0

    def __init__(self, dims, strides=None, name=None, scope=None, initializer=None, init_bias=0.0, padding='VALID', **kwargs):
        """
        :param dims: dims of filter 2d(height, width, channels), 1d(time, channels)
        """
        super(CNN, self).__init__(name, scope, initializer, **kwargs)
        self.dims = dims
        self.padding = padding
        self.init_bias = init_bias
        self.strides = strides

    def _create_graph(self, inputs):
        input_dim = inputs.get_shape().dims[-1].value
        dims = list(self.dims).copy()
        dims.insert(-1, input_dim)
        filters = tf.get_variable(self.name, dims, dtype=inputs.dtype)
        bias = tf.get_variable(self.name+'_b', dtype=inputs.dtype, initializer=tf.constant([self.init_bias]*self.dims[-1]))
        strides = self.strides
        if len(self.dims) == 3:
            if strides is None:
                strides = [1]*4
            outputs = tf.nn.conv2d(inputs, filters, strides, padding=self.padding) + bias
        elif len(self.dims) == 2:
            if strides is None:
                strides = 1
            outputs = tf.nn.conv1d(inputs, filters, strides, padding=self.padding) + bias
        return outputs


class RNN(Layer):
    uid = 0

    def __init__(self, dims, bidirection=True, cell_cls='tensorflow.nn.rnn_cell.GRUCell', dtype=tf.float32, swap_memory=True, name=None, scope=None, initializer=None, dropout=None, **kwargs):
        super(RNN, self).__init__(name, scope, initializer, **kwargs)
        self.dims = dims
        self.cell_cls = cell_cls
        self.bidirection = bidirection
        self.dtype = dtype
        self.dropout = dropout
        self.swap_memory = swap_memory
        self.cell = self._create_cell()
        if self.bidirection:
            self.cell_bw = self._create_cell()

    def _create_default_initializer(self):
        return tf.orthogonal_initializer()

    def _create_cell(self):
        multi_cell = tf_util.create_rnn_cell(self.dims, self.cell_cls, self.dropout)
        return multi_cell

    def _create_graph(self, inputs, inputs_lens=None, initial_state=None, rnn_func=None):
        if rnn_func is None:
            if self.bidirection:
                if initial_state is not None:
                    initial_state_fw, initial_state_bw = initial_state
                else:
                    initial_state_fw = None
                    initial_state_bw = None
                outs, stats = tf.nn.bidirectional_dynamic_rnn(self.cell, cell_bw, inputs, sequence_length = inputs_lens,  swap_memory = self.swap_memory,
                                                              dtype=self.dtype, initial_state_fw=initial_state_fw, initial_state_bw=initial_state_bw, scope=self.scope)
            else:
                outs, stats= tf.nn.dynamic_rnn(self.cell, inputs, sequence_length=inputs_lens,  swap_memory=self.swap_memory, dtype=self.dtype, initial_state=initial_state, scope=self.scope)
            if len(self.dims) == 1:
                if self.bidirection:
                    stats = tuple([s[-1] for s in stats])
                else:
                    stats = stats[-1]
            return outs, stats
        else:
            return rnn_func(self.cell, inputs, initial_state=initial_state)


class GumbelTeacherForceHelper(tf.contrib.seq2seq.TrainingHelper):
    def __init__(self, inputs, proj_layer, embedding, max_len, softmax_temperature, dropout=None):
        self._proj_layer = proj_layer
        self._embedding = embedding
        self._softmax_temperature = softmax_temperature
        self._dropout = dropout
        batch_size = tf.shape(inputs)[0]
        self._inputs_dim = inputs.get_shape().dims[-1]
        self._inputs_dtype = inputs.dtype
        sequence_lens = tf.ones([batch_size], dtype=tf.int32)*max_len
        super(GumbelTeacherForceHelper, self).__init__(inputs, sequence_lens)

    @property
    def sample_output_shape(self):
        return tuple([tensor_shape.TensorShape([self._inputs_dim]), tensor_shape.TensorShape([self._proj_layer.dim])])

    @property
    def sample_output_dtype(self):
        return self._inputs_dtype, self._inputs_dtype

    def sample(self, time, outputs, state, name=None):
        """sample for GreedyEmbeddingHelper."""
        del time, state  # unused by sample_fn
        # Outputs are logits, use argmax to get the most probable id
        if not isinstance(outputs, ops.Tensor):
            raise TypeError("Expected outputs to be a single Tensor, got: %s" %
                            type(outputs))
        logits = self._proj_layer(outputs)
        prob = tf_util.gumbel_softmax(logits, self._softmax_temperature)
        inp = tf.matmul(prob, self._embedding)
        if self._dropout is not None:
            inp = add_dropout(inp, dropout=self._dropout)
        return inp, logits

    def next_inputs(self, time, outputs, state, sample_outputs, name=None, **unused_kwargs):
        """next_inputs_fn for TrainingHelper."""
        with ops.name_scope(name, "TrainingHelperNextInputs",
                            [time, outputs, state]):
            next_time = time + 1
            finished = (next_time >= self._sequence_length)
            all_finished = math_ops.reduce_all(finished)
            next_inputs = sample_outputs[0]

            next_inputs = control_flow_ops.cond(
                all_finished, lambda: self._zero_inputs,
                lambda: next_inputs)
        return (finished, next_inputs, state)


class BasicDecoder(tf.contrib.seq2seq.BasicDecoder):
    class BasicDecoderOutput(namedtuple("BasicDecoderOutput", ("rnn_output", "sample_output"))):
        pass

    @property
    def output_size(self):
        # Return the cell output and the id
        if hasattr(self._helper, 'sample_output_shape'):
            sample_output = self._helper.sample_output_shape
        else:
            sample_output = self._helper.sample_ids_shape

        return self.BasicDecoderOutput(
            rnn_output=self._rnn_output_size(),
            sample_output=sample_output)

    @property
    def output_dtype(self):
        # Assume the dtype of the cell is the output_size structure
        # containing the input_state's first component's dtype.
        dtype = nest.flatten(self._initial_state)[0].dtype

        if hasattr(self._helper, 'sample_output_dtype'):
            sample_output = self._helper.sample_output_dtype
        else:
            sample_output = self._helper.sample_ids_dtype

        return self.BasicDecoderOutput(
            nest.map_structure(lambda _: dtype, self._rnn_output_size()),
            sample_output)

    def step(self, time, inputs, state, name=None):
        """Perform a decoding step.
        Args:
          time: scalar `int32` tensor.
          inputs: A (structure of) input tensors.
          state: A (structure of) state tensors and TensorArrays.
          name: Name scope for any created operations.
        Returns:
          `(outputs, next_state, next_inputs, finished)`.
        """
        with ops.name_scope(name, "BasicDecoderStep", (time, inputs, state)):
            cell_outputs, cell_state = self._cell(inputs, state)
            if self._output_layer is not None:
                cell_outputs = self._output_layer(cell_outputs)
            sample_outputs = self._helper.sample(
                time=time, outputs=cell_outputs, state=cell_state)
            if 'sample_outputs' in inspect.signature(self._helper.next_inputs).parameters:
                (finished, next_inputs, next_state) = self._helper.next_inputs(
                time=time,
                outputs=cell_outputs,
                state=cell_state,
                sample_outputs=sample_outputs)
            else:
                (finished, next_inputs, next_state) = self._helper.next_inputs(
                    time=time,
                    outputs=cell_outputs,
                    state=cell_state,
                    sample_ids=sample_outputs)
        outputs = self.BasicDecoderOutput(cell_outputs, sample_outputs)

        return (outputs, next_state, next_inputs, finished)


class DRNN(RNN):
    def _create_graph(self, inputs, inputs_lens=None, initial_state=None, max_len=None, helper_func=None, decoder_func=None, parallel_iterations=32, output_layer=None):
        if helper_func is None:
            return super(DRNN, self)._create_graph(inputs, inputs_lens, initial_state, None)
        else:
            dtype = inputs.dtype
            helper = helper_func(inputs)
            if initial_state is None:
                initial_state = self.cell.zero_state(tf.shape(inputs)[0], dtype)
            decoder = BasicDecoder(self.cell, helper, initial_state, output_layer=output_layer)
            outputs, state, sequence_lengths = tf.contrib.seq2seq.dynamic_decode(decoder, maximum_iterations=max_len, swap_memory=self.swap_memory, parallel_iterations=parallel_iterations, scope=self.scope)
            return outputs, state, sequence_lengths


class CudnnGRU(RNN):
    uid = 0
    def _create_graph(self, inputs, inputs_lens=None, initial_state=None, fetch_last_cell=True):
        if self.bidirection:
            direction = 'bidirectional'
        else:
            direction = 'unidirectional'
        rnn = tf.contrib.cudnn_rnn.CudnnGRU(len(self.dims), self.dims[-1], direction=direction)

        inputs = tf_util.transpose(inputs, 0, 1)
        outs, stats = rnn(inputs)
        stats = tuple(tf.squeeze(s, 0) for s in stats)
        outs = tf_util.transpose(outs, 0, 1)
        if fetch_last_cell:
            stats = stats[-1]
            if inputs_lens is not None:
                inds = tf.stack([tf.range(tf.shape(inputs_lens)[0]), inputs_lens - 1], 1)
                stats = tf.gather_nd(outs, inds)


        return outs, stats





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import logging
#from tensorflow.python.ops import array_ops, control_flow_ops, math_ops


from tensorflow.contrib import layers
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import rnn
from tensorflow.python.ops import tensor_array_ops
from tensorflow.python.ops import variable_scope as vs

__all__ = ["dynamic_rnn_decoder"]

def dynamic_rnn_decoder(cell, decoder_fn, inputs=None, sequence_length=None,
                        parallel_iterations=None, swap_memory=False,
                        time_major=False, scope=None, name=None):
  """ Dynamic RNN decoder for a sequence-to-sequence model specified by
  RNNCell and decoder function.
  The `dynamic_rnn_decoder` is similar to the `tf.python.ops.rnn.dynamic_rnn`
  as the decoder does not make any assumptions of sequence length and batch
  size of the input.
  The `dynamic_rnn_decoder` has two modes: training or inference and expects
  the user to create seperate functions for each.
  Under both training and inference, both `cell` and `decoder_fn` are expected,
  where `cell` performs computation at every timestep using `raw_rnn`, and
  `decoder_fn` allows modeling of early stopping, output, state, and next
  input and context.
  When training the user is expected to supply `inputs`. At every time step a
  slice of the supplied input is fed to the `decoder_fn`, which modifies and
  returns the input for the next time step.
  `sequence_length` is needed at training time, i.e., when `inputs` is not
  None, for dynamic unrolling. At test time, when `inputs` is None,
  `sequence_length` is not needed.
  Under inference `inputs` is expected to be `None` and the input is inferred
  solely from the `decoder_fn`.
  Args:
    cell: An instance of RNNCell.
    decoder_fn: A function that takes time, cell state, cell input,
      cell output and context state. It returns a early stopping vector,
      cell state, next input, cell output and context state.
      Examples of decoder_fn can be found in the decoder_fn.py folder.
    inputs: The inputs for decoding (embedded format).
      If `time_major == False` (default), this must be a `Tensor` of shape:
        `[batch_size, max_time, ...]`.
      If `time_major == True`, this must be a `Tensor` of shape:
        `[max_time, batch_size, ...]`.
      The input to `cell` at each time step will be a `Tensor` with dimensions
        `[batch_size, ...]`.
    sequence_length: (optional) An int32/int64 vector sized `[batch_size]`.
      if `inputs` is not None and `sequence_length` is None it is inferred
      from the `inputs` as the maximal possible sequence length.
    parallel_iterations: (Default: 32).  The number of iterations to run in
      parallel.  Those operations which do not have any temporal dependency
      and can be run in parallel, will be.  This parameter trades off
      time for space.  Values >> 1 use more memory but take less time,
      while smaller values use less memory but computations take longer.
    swap_memory: Transparently swap the tensors produced in forward inference
      but needed for back prop from GPU to CPU.  This allows training RNNs
      which would typically not fit on a single GPU, with very minimal (or no)
      performance penalty.
    time_major: The shape format of the `inputs` and `outputs` Tensors.
      If true, these `Tensors` must be shaped `[max_time, batch_size, depth]`.
      If false, these `Tensors` must be shaped `[batch_size, max_time, depth]`.
      Using `time_major = True` is a bit more efficient because it avoids
      transposes at the beginning and end of the RNN calculation.  However,
      most TensorFlow data is batch-major, so by default this function
      accepts input and emits output in batch-major form.
    scope: VariableScope for the `raw_rnn`;
      defaults to None.
    name: NameScope for the decoder;
      defaults to "dynamic_rnn_decoder"
  Returns:
    A tuple (outputs, final_state, final_context_state) where:
      outputs: the RNN output 'Tensor'.
        If time_major == False (default), this will be a `Tensor` shaped:
          `[batch_size, max_time, cell.output_size]`.
        If time_major == True, this will be a `Tensor` shaped:
          `[max_time, batch_size, cell.output_size]`.
      final_state: The final state and will be shaped
        `[batch_size, cell.state_size]`.
      final_context_state: The context state returned by the final call
        to decoder_fn. This is useful if the context state maintains internal
        data which is required after the graph is run.
        For example, one way to diversify the inference output is to use
        a stochastic decoder_fn, in which case one would want to store the
        decoded outputs, not just the RNN outputs. This can be done by
        maintaining a TensorArray in context_state and storing the decoded
        output of each iteration therein.
  Raises:
    ValueError: if inputs is not None and has less than three dimensions.
  """
  with ops.name_scope(name, "dynamic_rnn_decoder",
                      [cell, decoder_fn, inputs, sequence_length,
                       parallel_iterations, swap_memory, time_major, scope]):
    if inputs is not None:
      # Convert to tensor
      inputs = ops.convert_to_tensor(inputs)

      # Test input dimensions
      if inputs.get_shape().ndims is not None and (
          inputs.get_shape().ndims < 2):
        raise ValueError("Inputs must have at least two dimensions")
      # Setup of RNN (dimensions, sizes, length, initial state, dtype)
      if not time_major:
        # [batch, seq, features] -> [seq, batch, features]
        inputs = array_ops.transpose(inputs, perm=[1, 0, 2])

      dtype = inputs.dtype
      # Get data input information
      input_depth = int(inputs.get_shape()[2])
      batch_depth = inputs.get_shape()[1].value
      max_time = inputs.get_shape()[0].value
      if max_time is None:
        max_time = array_ops.shape(inputs)[0]
      # Setup decoder inputs as TensorArray
      inputs_ta = tensor_array_ops.TensorArray(dtype, size=max_time)
      inputs_ta = inputs_ta.unstack(inputs)

    def loop_fn(time, cell_output, cell_state, loop_state):
      if cell_state is None:  # first call, before while loop (in raw_rnn)
        if cell_output is not None:
          raise ValueError("Expected cell_output to be None when cell_state "
                           "is None, but saw: %s" % cell_output)
        if loop_state is not None:
          raise ValueError("Expected loop_state to be None when cell_state "
                           "is None, but saw: %s" % loop_state)
        context_state = None
      else:  # subsequent calls, inside while loop, after cell excution
        if isinstance(loop_state, tuple):
          (done, context_state) = loop_state
        else:
          done = loop_state
          context_state = None

      # call decoder function
      if inputs is not None:  # training
        # get next_cell_input
        if cell_state is None:
          next_cell_input = inputs_ta.read(0)
        else:
          if batch_depth is not None:
            batch_size = batch_depth
          else:
            batch_size = array_ops.shape(done)[0]
          next_cell_input = control_flow_ops.cond(
              math_ops.greater(time, max_time-1),
              lambda: array_ops.zeros([batch_size, input_depth], dtype=dtype),
              lambda: inputs_ta.read(time))
        (next_done, next_cell_state, next_cell_input, emit_output,
         next_context_state) = decoder_fn(time, cell_state, next_cell_input,
                                          cell_output, context_state)
      else:  # inference
        # next_cell_input is obtained through decoder_fn
        (next_done, next_cell_state, next_cell_input, emit_output,
         next_context_state) = decoder_fn(time, cell_state, None, cell_output,
                                          context_state)

      # check if we are done
      if next_done is None:  # training
        next_done = time >= sequence_length

      # build next_loop_state
      if next_context_state is None:
        next_loop_state = next_done
      else:
        next_loop_state = (next_done, next_context_state)

      return (next_done, next_cell_input, next_cell_state,
              emit_output, next_loop_state)

    # Run raw_rnn function
    outputs_ta, final_state, final_loop_state = rnn.raw_rnn(
        cell, loop_fn, parallel_iterations=parallel_iterations,
        swap_memory=swap_memory, scope=scope)
    outputs = outputs_ta.stack()

    # Get final context_state, if generated by user
    if isinstance(final_loop_state, tuple):
      final_context_state = final_loop_state[1]
    else:
      final_context_state = None

    if not time_major:
      # [seq, batch, features] -> [batch, seq, features]
      outputs = array_ops.transpose(outputs, perm=[1, 0, 2])
    return outputs, final_state, final_context_state


def gan_decoder(init_state, embedding, projecting, softmax, maximum_length, SOS, EOS, start_idx =tf.constant(0), name = None):
    batch_size = tf.shape(init_state[0][0])[0]; zero = tf.constant(0)
    context_ta = tf.TensorArray(dtype=tf.float32, size=maximum_length, dynamic_size=False, infer_shape=True)
    def decoder_fn(time, cell_state, cell_input, cell_output, context_state):
        def sample():
            output_logits = projecting(cell_output)
            output_prob = softmax(output_logits) + 1e-14
            output_log_prob = tf.log(output_prob)
            output_ids = tf.reshape(tf.multinomial(output_log_prob, 1), [batch_size])
            output_sample_log_prob = tf.gather_nd(output_log_prob, tf.stack([tf.range(batch_size), tf.cast(output_ids,tf.int32)], 1))
            next_cell_input = embedding(output_ids)
            next_done = tf.equal(output_ids, EOS)
            next_context_state = context_state.write(time-start_idx-1, tf.stack([output_sample_log_prob, tf.cast(output_ids, tf.float32)]))
            return next_cell_input, next_done, next_context_state
        def teach():
            return cell_input, tf.zeros([batch_size], dtype = tf.bool), context_ta
        def embed_fn():
            embeded = embedding(tf.ones([batch_size], dtype=tf.int32) * SOS)
            return embeded
            
         
        if cell_output is None:
            next_done = tf.zeros([batch_size], dtype = tf.bool)

           # next_cell_input = control_flow_ops.cond(tf.greater(start_idx, zero), embed_fn, embed_fn)
            next_cell_input = embedding(tf.ones([batch_size], dtype=tf.int32) * SOS)
                
            next_cell_state = init_state
            next_cell_output = cell_output
            next_context_state = context_ta
        elif cell_input is None:
            #cond = tf.logical_and(tf.greater(start_idx, time-1), cell_input is not None)
            next_cell_input, next_done, next_context_state = sample()
            next_cell_state = cell_state
            next_cell_output = cell_output
        else:
            next_cell_input, next_done, next_context_state = control_flow_ops.cond(tf.greater(start_idx, time-1), teach, sample)
            next_cell_state = cell_state
            next_cell_output = cell_output
            
        next_done = control_flow_ops.cond(math_ops.greater(time, maximum_length - 1), lambda: array_ops.ones([batch_size, ], dtype=tf.bool), lambda: next_done)
        return next_done, next_cell_state, next_cell_input, next_cell_output, next_context_state
    return decoder_fn




if __name__ == "__main__":
    from Model import BasicModel as Model
    import numpy as np, sys
    from layers import *
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG, format='%(asctime)s %(levelname)s:%(message)s')

    batch_size = 2; dec_dims = 8; embed_size = 4; embed_dim = 8; maximum_length = 100; start_idx = 0
    embed = np.random.rand(embed_size, embed_dim).astype(np.float32)
    embed[0,:] = 0.5; embed[1,:] = 1 
    
    embedding_fn = embedding_layer(embed_size, embed_dim, embedding = embed, trainable = False)
    projecting_fn = dense_layer(embed_size, name = "projecting")
    aaa  = embedding_fn(tf.ones([batch_size], dtype=tf.int32) * 1)
    bbb = projecting_fn(tf.constant(np.random.rand(2,8), tf.float32))
    softmax_fn = softmax_layer()
    decoder = Model._create_decoder([dec_dims, dec_dims])
    state = decoder.zero_state(batch_size);
    decoder_fn = gan_decoder(state, embedding_fn, projecting_fn, softmax_fn, maximum_length, 0, 1, start_idx )
    vals = np.random.rand(batch_size, 9, dec_dims)
    vals[:,0,:] = 0.5
    vals[:,8,:] = 1
    print(embed)
    inputs = tf.constant(vals, tf.float32)
    outputs, state, context_state = decoder(inputs[:,0:-1:,:], decoder_fn = decoder_fn)
    #max_len =tf.shape(outputs)[1]
    max_len = tf.reduce_max(tf.reduce_sum(tf.cast(tf.greater(tf.reduce_sum(outputs, -1),0.0), tf.int32), -1)) 
    context = tf.transpose(context_state.gather(tf.range(max_len - start_idx )), [1,2,0])
    logprob = tf.squeeze(tf.slice(context,[0, 0, 0], [1, -1, -1]), [0])
    ids = tf.squeeze(tf.slice(context,[1, 0, 0], [1, -1, -1]), [0])
    gradients = tf.gradients(logprob, tf.trainable_variables())
    with tf.Session() as sess:
        for i in range(1):
            sess.run(tf.global_variables_initializer())
            output, logprob, ids = sess.run([outputs, logprob, ids])
            logging.info("outputs shape is:{}, output is: {}, ids:{} ".format(output.shape, output, ids))
            logging.info("logprob shape is {}, logprob is: {}".format(logprob.shape, logprob))
            output = sess.run(gradients)
            logging.info("gradients len is:{}, output is: {}".format(len(output), output[0]))

            
    



import tensorflow as tf
import numpy as np

class RVQ(tf.keras.layers.Layer):
    def __init__(self, k, d, h, **kwargs):
        """
        :param k: number of vector in the code book
        :param d: dimension of vector
        :param h: num of heads
        """
        self.k = k
        self.d = d
        self.h = h
        m = np.log2(k)  # m bits
        assert m.is_integer()
        assert m//h == 0

        self.k_ = int(np.power(2, m//h))
        self.kernel = self.add_variable("kernel", shape=[h, self.k_, d//h])
        self.alpha = self.add_variable('alpha')

    def dist(self, input):
        repeats = [1] * input.shape.ndims
        repeats.insert(-1, self.h)
        repeats.insert(-1, self.k_)

        src = tf.expand_dims(input, -2)
        src = tf.expand_dims(src, -2)
        src = tf.tile(src, repeats)
        tgt = self.kernel
        for _ in repeats[0:-3]:
            tgt = tf.expand_dims(tgt, 0)
        d = tf.norm(src - tgt, axis=-1)
        return d

    def __call__(self, input):
        d = self.dist(input)  # [::, h, k_]
        ind = tf.argmin(d, -1)  # [::, h]
        v = tf.gather(self.kernel, ind)  # [::, h, d//h]
        shape = tf.shape(v)
        new_shape = tf.concat([shape[:-2], [self.d]], axis=0)
        v = tf.reshape(v, new_shape)
        v = self.alpha * input + (1 - self.alpha) * v
        base = tf.pow(self.k_, tf.range(h))
        ind *= base
        ind = tf.reduce_sum(ind, -1)
        return v, ind


class RVQA(tf.keras.Model):
    """
    residual vector quantized auto encoder
    paper: Unsupervised Paraphrasing without Translation https://arxiv.org/pdf/1905.12752.pdf

    """
    def __init__(self, k, d, encoder, decoder, **kwargs):
        """
        :param k: number of vector in the code book
        :param d: dimension of vector
        """
        self.vq = VQ(k, d)
        self.encoder = encoder
        self.decoder = decoder

        super(RVQA, self).__init__(**kwargs)


    def __call__(self, input):
        z_e = self.encoder(input)
        z_q, z_ind = self.vq(z_e)




